local C = Growler_Constants
local O = Growler_Options
local S = Growler_Stable
local l = BixLib
local SL = Growler_Slash
local _,L = ...

local tinsert, tostring, print, pairs, SendAddonMessage, PlaySoundKitID, CreateFrame, ActionButton_ShowOverlayGlow, ActionButton_HideOverlayGlow
		= tinsert, tostring, print, pairs, SendAddonMessage, PlaySoundKitID, CreateFrame, ActionButton_ShowOverlayGlow, ActionButton_HideOverlayGlow

local class = tostring(UnitClass("player"))
local isHunter = (class == L["Hunter"])

local addonsLoaded = {}

local TimeSinceLastUpdate = 0
local TimeSinceLastPetAttack = 0
local TimeSinceLastPlayerAttack = 0
local TimeSinceLastScreenIconUpdate = 0
local TimeFalling = 0
local WaitForPet = 0
local IsJumping = false
local IsJumpingNotFalling = false
local TipPlayDead = false
local Growler_PetCalled = 4

local scale = 1
local siscale = 1
local btnPadding = C.btnPadding
local btnSpacing = C.btnSpacing
local barbtnSize = C.barbtnSize * scale
local btnSize = C.btnSize * scale
local petbtnSize = C.petbtnSize * scale
local stancebtnSize = C.stancebtnSize * scale
local petbtnOffset = C.petbtnOffset * scale
local siSize = C.siSize * siscale
local portraitSize = C.portraitSize

local Growler_InitiateDB = function()
	GrowlerDB = GrowlerDB or {}
	GrowlerDB.options = GrowlerDB.options or {}
	
	Growler.DefaultDB = {}
	Growler.DefaultDB.options = {}
	for c,cat in pairs(C.Options) do
		for o,opt in pairs(C.Options[c]) do
			if GrowlerDB.options[o] == nil then
				GrowlerDB.options[o] = C.Options[c][o].default
			end
			Growler.DefaultDB.options[o] = C.Options[c][o].default
		end
	end
	
	GrowlerDB.pets = GrowlerDB.pets or {}
	Growler.DefaultDB.pets = Growler.DefaultDB.pets or {}
	for i=1,55 do
		Growler.DefaultDB.pets[i] = {}
		Growler.DefaultDB.pets[i].icon = ""
		Growler.DefaultDB.pets[i].name = ""
		Growler.DefaultDB.pets[i].level = ""
		Growler.DefaultDB.pets[i].family = ""
		Growler.DefaultDB.pets[i].talent = ""
		if GrowlerDB.pets[i] == nil then
			GrowlerDB.pets[i] = {}
			GrowlerDB.pets[i].icon = ""
			GrowlerDB.pets[i].name = ""
			GrowlerDB.pets[i].level = ""
			GrowlerDB.pets[i].family = ""
			GrowlerDB.pets[i].talent = ""
		end
	end
end

local Growler_Setup = function()
	if GrowlerDB.options["savepos"] then
		scale = GrowlerDB.options.barsize
		siscale = GrowlerDB.options.sisize
	else
		scale = Growler.DefaultDB.options.barsize
		siscale = Growler.DefaultDB.options.sisize
	end
	btnPadding = C.btnPadding
	btnSpacing = C.btnSpacing
	barbtnSize = C.barbtnSize * scale
	btnSize = C.btnSize * scale
	petbtnSize = C.petbtnSize * scale
	stancebtnSize = C.stancebtnSize * scale
	petbtnOffset = C.petbtnOffset * scale
	siSize = C.siSize * siscale
	portraitSize = C.portraitSize * scale
	
	Growler.btn = {}
	Growler.petbtn = {}
	Growler.specbtn = {}
	Growler.stancebtn = {}
	Growler.GrowlMessageSent = false
	Growler.GrowlWasCast = false
	Growler.LastMessageCheck = 0
	Growler.RecievedMessages = {}
	Growler.EnteredWorld = false
	Growler.PetAvailable = false
	Growler.RefreshKeybinds = true
	Growler.TalentOrButtonChanged = false
	Growler.TalentChangeMsgShown = true
	Growler.TalentChangeLastMsg = 10
	Growler.TalentChangeLastUpdate = 0
	
	Growler.HasVisitedStables = function()
		local emptyslots = 0
		for i=1,55 do
			if GrowlerDB.pets[i].icon == "" then
				emptyslots = emptyslots + 1
			end
		end
		
		return emptyslots < 55
	end
	
	
	Growler.SendAddonMsg = function(msg)
		SendAddonMessage("growler",msg,"PARTY")
		SendAddonMessage("growler",msg,"RAID")
	end
	
	Growler.AddonMsgRecieved = function(prefix,msg,channel,sender)
		if prefix == "growler" then
			sender = l.replace(sender,"-","_")
			Growler.RecievedMessages[sender] = msg		
		end
	end
	
	Growler.CheckForGroupGrowl = function(timestamp,event,_,sourceGUID,sourceName,sourceFlags,sourceRaidFlags,destGUID,destName,destFlags,destRaidFlags,spellID,spellName)
		if not Growler.GrowlMessageSent and IsInGroup() and GetNumGroupMembers() > 1 and (spellID == 2649 or spellID == 171014 or spellID == 17735) and (UnitPlayerOrPetInRaid(sourceName) or UnitPlayerOrPetInParty(sourceName)) then
			local inInstance, instanceType = IsInInstance()
			if inInstance then
			
				local maxPlayers = select(5,GetInstanceInfo())
				if instanceType ~= "arena" and instanceType ~= "pvp" and instanceType ~= "none" and maxPlayers > 1 then
					local isBoss = false
					local _, _, encountersTotal, encountersComplete = GetInstanceLockTimeRemaining()
					for i=1,encountersTotal do
						local bossName, texture, isKilled = GetInstanceLockTimeRemainingEncounter(i)
						if not isKilled and (destName == bossName) then
							isBoss = true
							break
						end
					end
					if not isBoss then
						Growler.SendAddonMsg("growl;true")
						Growler.GrowlWasCast = true
					end
				end
			end
		end
	end
	
	Growler.CheckCombatLogEvent = function(timestamp,event,_,sourceGUID,sourceName,sourceFlags,sourceRaidFlags,destGUID,destName,destFlags,destRaidFlags,spellID,spellName)
		if tostring(sourceGUID) == UnitGUID("pet") and event == "SWING_DAMAGE" then
			TimeSinceLastPetAttack = 0
		elseif tostring(sourceGUID) == UnitGUID("player") and (event == "SWING_DAMAGE" or event == "RANGE_DAMAGE" or event == "SPELL_DAMAGE") then
			TimeSinceLastPlayerAttack = 0
		end
	end
	
	Growler.msg = function(msg,warning,news)
		local color1 = "|cFFff4433"
		local color2 = "|cFFaa8822"
		local News = news and "news" or ""
		print(color1 .. "{{" .. color2 .. "Growler " .. News .. color1 .. "::" .. color2 .. " " .. msg .. color1 .. "}}")
		if warning then PlaySoundKitID(47355) end
		--if warning then PlaySoundKitID(34040) end
	end
	
	Growler.GetActivePetSlot = function()
		local curName,_ = UnitName("pet")
		local curFamily = UnitCreatureFamily("pet")
		local curIcon = GetPetIcon()
		local activepetslot = 0
		for i=1,5 do
			local petIcon, petName, _, petType, _ = GetStablePetInfo(i)
			if petIcon ~= nil and petIcon == curIcon and petName == curName and petType == curFamily then
				activepetslot = i
			end
		end
		return activepetslot
	end
				
	Growler.SetActivePetSlot = function(load)
		if not InCombatLockdown() then
			local curName,_ = UnitName("pet")
			local curFamily = UnitCreatureFamily("pet")
			local curIcon = GetPetIcon()
			local activepetslot = 0
			if load ~= nil then
				activepetslot = GrowlerDB.options.activepetslot
			else
				if curName ~= nil then
					for i=1,5 do
						local petIcon, petName, _, petType, _ = GetStablePetInfo(i)
						if petIcon ~= nil and petIcon == curIcon and petName == curName and petType == curFamily then
							activepetslot = i
						end
					end
					
					GrowlerDB.options.activepetslot = activepetslot
					
					Growler.btn.call:SetAttribute("macrotext1","/cast [pet] "..GetSpellInfo(982).."\n/castsequence [nopet] reset=3 "..GetSpellInfo(l.pets.callpet[activepetslot]) .. ", "..GetSpellInfo(982))
					Growler.msg(L["Active pet slot(%s) saved."]:format(activepetslot))
					PlaySoundKitID(13827)
					Growler.RefreshKeybinds = true
				else
					Growler.msg(L["Summon a pet to save that as your active slot."],true)
				end
			end
			if activepetslot == 0 and load then
				Growler.msg(L["You do not have an active pet slot saved. Summon a pet and click the save button."])
			end
		else
			if not load then
				Growler.msg(L["You cannot change your active pet in combat."],true)
			end
		end
	end
	
	Growler.UpdateSI = function()
		local prior = 0
		local btn = nil
		for k,v in pairs(Growler.btn) do
			if v.active and v.priority > prior then
				prior = v.priority
				btn = v
			end
		end
		for k,v in pairs(Growler.btn) do
			if not v.active or v.priority < prior then
				v.si.hide()
			end
		end
		if prior > 0 and GrowlerDB.options.screenicons then
			if TimeSinceLastScreenIconUpdate > 2.5 then
				if btn.si.active then
					btn.si:SetAlpha(0.3)
					btn.si:SetSize(siSize*0.9,siSize*0.9)
					btn.si.active = false
				else
					btn.si:SetAlpha(0.5)
					btn.si:SetSize(siSize,siSize)
					btn.si.active = true
				end
				TimeSinceLastScreenIconUpdate = 0
			end
		else
			TimeSinceLastScreenIconUpdate = 0
			for k,v in pairs(Growler.btn) do
				v.si.hide()
			end
		end
	end
	
	Growler.CreateSI = function(name,icon)
		local siName = "GrowlerSI_" .. name
		local si = CreateFrame("FRAME", siName, GrowlerBar)
		si.name = siName
		si.lastupdate = 0
		si.active = false
		si:SetSize(siSize,siSize)
		l.SetTexture(icon,si)
		si:SetAlpha(0)
		si.show = function()
			if GrowlerDB.options.screenicons then
				TimeSinceLastScreenIconUpdate = 0
				si:SetAlpha(0.5)
				si:SetSize(siSize,siSize)
				si.active = true
			end
		end
		si.hide = function()
			si.active = false
			si:SetAlpha(0)
		end
		
		return si
	end
	
	Growler.CreateSIPosFrame = function()
		local siName = "GrowlerSI_PosFrame"
		local si = getglobal("GrowlerSI_PosFrame") or CreateFrame("BUTTON", siName, UIParent)
		si.name = siName
		si:SetSize(siSize,siSize)
		l.SetTexture("Interface\\ICONS\\ability_physical_taunt",si)
		si:SetAlpha(1)
		si:Hide()
		si:SetMovable(false)
		si:SetClampedToScreen(true)
		
		si:RegisterForDrag("LeftButton")
		si:SetScript("OnDragStart",function(self)
			si:SetMovable(true)
			si:StartMoving()
		end)
		si:SetScript("OnDragStop",function(self)
			si:StopMovingOrSizing()
			si:SetMovable(false)
			Growler.bar.SavePos()
		end)
		
		return si
	end
	Growler.CreateSIPosFrame()
		
	Growler.CreateBar = function()
		Growler.bar = CreateFrame("BUTTON", "GrowlerBar", UIParent,"SecureHandlerStateTemplate")
		Growler.bar.movable = false
		Growler.bar:SetAttribute("_onstate-busy", [[ -- arguments: self, stateid, newstate
			if newstate ~= "no" then
				self:Hide()
			else
				self:Show()
			end
			
			if newstate == "pet" then
				self:ClearBindings()
			end
		]])
		RegisterStateDriver(GrowlerBar, "busy", "[petbattle] pet; [shapeshift][bonusbar][vehicleui][possessbar][overridebar][flying] yes; no")
		-- Growler.bar:SetAttribute("_onstate-test", [[ -- arguments: self, stateid, newstate
			-- print(newstate)
		-- ]])
		-- RegisterStateDriver(GrowlerBar, "test", "[shapeshift] shapeshift; [bonusbar] bonusbar; [vehicleui] vehicleui; [possessbar] possessbar; [overridebar] overridebar; none")
		Growler.bar:SetScript("OnShow",function(self)
			Growler.RefreshKeybinds = true
		end)
		Growler.bar:SetScript("OnHide",function(self)
			for k,btn in pairs(Growler.btn) do
				btn.hide()
				btn.si.hide()
			end
		end)
		Growler.bar:SetMovable(false)
		Growler.bar:SetClampedToScreen(true)
		Growler.bar:RegisterForClicks("AnyDown")
		Growler.bar:SetScript("OnMouseDown", function(self,button) Growler.bar.MouseDown(self,button) end)
		Growler.bar:RegisterForDrag("LeftButton")
		Growler.bar:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
		Growler.bar:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
		Growler.bar.DragStart = function(self)
			if Growler.bar.movable then
				Growler.bar:SetMovable(true)
				Growler.bar:StartMoving()
			end
		end
		Growler.bar.DragStop = function(self)
			Growler.bar:StopMovingOrSizing()
			Growler.bar:SetMovable(false)
			Growler.bar.SavePos()
		end
		Growler.bar.MouseDown = function(self, button)
			if button == "RightButton" then
			elseif button == "MiddleButton" then
			end
		end
		local backdrop = {
		  bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
		  -- bgFile = "Interface\\FrameGeneral\\UI-Background-Rock",
		  edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
		  tile = true,
		  tileSize = 2*btnPadding+btnSize,
		  edgeSize = 16,
		  insets = {left = 5,right = 5,top = 5,bottom = 5}
		}
		Growler.bar:SetBackdrop(backdrop)
		Growler.bar:SetBackdropColor(1,1,1,1)
		Growler.bar:SetBackdropBorderColor(1,1,1,1)
		
		
		Growler.bar.resetbtn = CreateFrame("BUTTON","GrowlerBar_PetSlotButton",Growler.bar)
		Growler.bar.resetbtn.size = 15 * scale
		Growler.bar.resetbtn:SetSize(Growler.bar.resetbtn.size,Growler.bar.resetbtn.size-1)
		Growler.bar.resetbtn:SetPoint("TOPRIGHT",Growler.bar,"TOPRIGHT",(-5),(-8))
		Growler.bar.resetbtn.texture_up = "Interface\\WorldMap\\X_Mark_64Grey"
		Growler.bar.resetbtn.texture_down = "Interface\\WorldMap\\X_Mark_64Grey"
		l.SetTexture(Growler.bar.resetbtn.texture_up,Growler.bar.resetbtn)
		Growler.bar.resetbtn:RegisterForDrag("LeftButton")
		Growler.bar.resetbtn:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
		Growler.bar.resetbtn:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
		Growler.bar.resetbtn:RegisterForClicks("LeftButtonUp","LeftButtonDown")
		Growler.bar.resetbtn:SetScript("OnClick",function(self,button,down)
			if button == "LeftButton" then
				if down then
					Growler.bar.resetbtn.texture:SetTexture(Growler.bar.resetbtn.texture_down)
				else
					Growler.bar.resetbtn.texture:SetTexture(Growler.bar.resetbtn.texture_up)
					Growler.bar.ResetPos(true)
					Growler.bar.SavePos()
				end
			end
		end)
		Growler.bar.resetbtn.update = function()
		end
		Growler.bar.resetbtn.setTooltipText = function()
			local tt = L["%sReset bar position."]:format(C.color.white)
			GameTooltip:SetText(tt)
		end
		Growler.bar.resetbtn.updateTooltip = function()
			if Growler.bar.resetbtn.mouseover then
				Growler.bar.resetbtn.setTooltipText()
			end
		end
		Growler.bar.resetbtn:SetScript("OnEnter",function(self)
			Growler.bar.resetbtn.mouseover = true
			GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
			Growler.bar.resetbtn.setTooltipText()
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("BOTTOMLEFT",Growler.bar.resetbtn,"TOPLEFT",0,0)
			GameTooltip:Show()
			Growler.bar.resetbtn.updateTooltip()
		end)
		Growler.bar.resetbtn:SetScript("OnLeave",function(self)
			Growler.bar.resetbtn.mouseover = false
			GameTooltip:ClearAllPoints()
			GameTooltip:Hide()
		end)
		
		Growler.bar.settingsbtn = CreateFrame("BUTTON","GrowlerBar_PetSlotButton",Growler.bar)
		Growler.bar.settingsbtn.size = 15 * scale
		Growler.bar.settingsbtn:SetSize(Growler.bar.settingsbtn.size,Growler.bar.settingsbtn.size)
		Growler.bar.settingsbtn:SetPoint("BOTTOMRIGHT",Growler.bar,"BOTTOMRIGHT",(-5),10)
		Growler.bar.settingsbtn.texture_up = "Interface\\WorldMap\\Gear_64Grey"
		Growler.bar.settingsbtn.texture_down = "Interface\\WorldMap\\Gear_64Grey"
		l.SetTexture(Growler.bar.settingsbtn.texture_up,Growler.bar.settingsbtn)
		Growler.bar.settingsbtn:RegisterForDrag("LeftButton")
		Growler.bar.settingsbtn:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
		Growler.bar.settingsbtn:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
		Growler.bar.settingsbtn:RegisterForClicks("LeftButtonUp","LeftButtonDown")
		Growler.bar.settingsbtn:SetScript("OnClick",function(self,button,down)
			if button == "LeftButton" then
				if down then
					Growler.bar.settingsbtn.texture:SetTexture(Growler.bar.settingsbtn.texture_down)
				else
					Growler.bar.settingsbtn.texture:SetTexture(Growler.bar.settingsbtn.texture_up)
					InterfaceOptionsFrame_OpenToCategory(O)
					InterfaceOptionsFrame_OpenToCategory(O)
				end
			end
		end)
		Growler.bar.settingsbtn.update = function()
		end
		Growler.bar.settingsbtn.setTooltipText = function()
			local tt = L["%sSettings."]:format(C.color.white)
			GameTooltip:SetText(tt)
		end
		Growler.bar.settingsbtn.updateTooltip = function()
			if Growler.bar.settingsbtn.mouseover then
				Growler.bar.settingsbtn.setTooltipText()
			end
		end
		Growler.bar.settingsbtn:SetScript("OnEnter",function(self)
			Growler.bar.settingsbtn.mouseover = true
			GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
			Growler.bar.settingsbtn.setTooltipText()
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("BOTTOMLEFT",Growler.bar.settingsbtn,"TOPLEFT",0,0)
			GameTooltip:Show()
			Growler.bar.settingsbtn.updateTooltip()
		end)
		Growler.bar.settingsbtn:SetScript("OnLeave",function(self)
			Growler.bar.settingsbtn.mouseover = false
			GameTooltip:ClearAllPoints()
			GameTooltip:Hide()
		end)
		
		Growler.bar.lockbtn = CreateFrame("BUTTON","GrowlerBar_Lockbutton",Growler.bar)
		Growler.bar.lockbtn.pos = 1
		Growler.bar.lockbtn.size = barbtnSize
		Growler.bar.lockbtn:SetSize(Growler.bar.lockbtn.size,btnSize)
		Growler.bar.lockbtn:SetPoint("RIGHT",Growler.bar,"RIGHT",((-Growler.bar.settingsbtn.size) - btnSpacing) * Growler.bar.lockbtn.pos,0 * scale)
		Growler.bar.lockbtn.texture_up = "Interface\\BUTTONS\\UI-Button-KeyRing"
		Growler.bar.lockbtn.texture_down = "Interface\\BUTTONS\\UI-Button-KeyRing-Down"
		Growler.bar.lockbtn.texture_disabled = ""
		l.SetTexture(Growler.bar.lockbtn.texture_down,Growler.bar.lockbtn)
		Growler.bar.lockbtn.texture:SetPoint("TOPLEFT",Growler.bar.lockbtn,"TOPLEFT",0 * scale,0 * scale)
		Growler.bar.lockbtn.texture:SetPoint("BOTTOMRIGHT",Growler.bar.lockbtn,"BOTTOMRIGHT",16 * scale,(-18) * scale)
		Growler.bar.lockbtn:RegisterForDrag("LeftButton")
		Growler.bar.lockbtn:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
		Growler.bar.lockbtn:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
		Growler.bar.lockbtn:RegisterForClicks("LeftButtonUp")
		Growler.bar.lockbtn:SetScript("OnMouseUp",function(self,button)
			if button == "LeftButton" then
				PlaySoundKitID(8939)
				if Growler.bar.movable then 
					Growler.bar:SetBackdropBorderColor(1,1,1,1)	
					Growler.bar.lockbtn.texture:SetTexture(Growler.bar.lockbtn.texture_down)
					Growler.bar.movable = false
					GrowlerSI_PosFrame:Hide()
				else
					Growler.bar:SetBackdropBorderColor(0.7,1,0.7,1)
					Growler.bar.lockbtn.texture:SetTexture(Growler.bar.lockbtn.texture_up)
					Growler.bar.movable = true
					GrowlerSI_PosFrame:Show()
				end
				Growler.bar.lockbtn.setTooltipText()
			end
		end)
		Growler.bar.lockbtn.update = function()
		end
		Growler.bar.lockbtn.setTooltipText = function()
			local tt = L["%sLock %sthe bar from dragging."]:format("|cffff0000",C.color.white)
			if not Growler.bar.movable then
				tt = L["%sUnlock %sthe bar for dragging."]:format("|cff00ff00",C.color.white)
			end
			GameTooltip:SetText(tt)
		end
		Growler.bar.lockbtn.updateTooltip = function()
			if Growler.bar.lockbtn.mouseover then
				Growler.bar.lockbtn.setTooltipText()
			end
		end
		Growler.bar.lockbtn:SetScript("OnEnter",function(self)
			Growler.bar.lockbtn.mouseover = true
			GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
			Growler.bar.lockbtn.setTooltipText()
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("BOTTOMLEFT",Growler.bar.lockbtn,"TOPLEFT",0,0)
			GameTooltip:Show()
			Growler.bar.lockbtn.updateTooltip()
		end)
		Growler.bar.lockbtn:SetScript("OnLeave",function(self)
			Growler.bar.lockbtn.mouseover = false
			GameTooltip:ClearAllPoints()
			GameTooltip:Hide()
		end)
		
	
		Growler.bar.stablebtn = CreateFrame("BUTTON","GrowlerBar_Lockbutton",Growler.bar)
		Growler.bar.stablebtn.pos = 2
		Growler.bar.stablebtn:SetSize(barbtnSize,btnSize)
		Growler.bar.stablebtn:SetPoint("RIGHT",Growler.bar,"RIGHT",((-Growler.bar.settingsbtn.size) - btnSpacing) * Growler.bar.stablebtn.pos,0 * scale)
		Growler.bar.stablebtn.texture_up = "Interface\\BUTTONS\\UI-MicroButton-Spellbook-Up"
		Growler.bar.stablebtn.texture_down = "Interface\\BUTTONS\\UI-MicroButton-Spellbook-Down"
		Growler.bar.stablebtn.texture_disabled = "Interface\\BUTTONS\\UI-MicroButton-Spellbook-Disabled"
		l.SetTexture(Growler.bar.stablebtn.texture_up,Growler.bar.stablebtn)
		Growler.bar.stablebtn.texture:SetPoint("TOPLEFT",Growler.bar.stablebtn,"TOPLEFT",0 * scale,20 * scale)
		Growler.bar.stablebtn.texture:SetPoint("BOTTOMRIGHT",Growler.bar.stablebtn,"BOTTOMRIGHT",0 * scale,(-2) * scale)
		Growler.bar.stablebtn:RegisterForDrag("LeftButton")
		Growler.bar.stablebtn:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
		Growler.bar.stablebtn:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
		Growler.bar.stablebtn:RegisterForClicks("LeftButtonUp")
		Growler.bar.stablebtn:SetScript("OnMouseUp",function(self,button)
			if button == "LeftButton" then
				if not Growler.HasVisitedStables() then
					Growler.msg(L["Visit the stables to use the stable function. (you only have to do this once)"],true)
				else
					if S.IsOpen then
						Growler.bar.stablebtn.texture:SetTexture(Growler.bar.stablebtn.texture_up)
						S.HideStable()
					else
						Growler.bar.stablebtn.texture:SetTexture(Growler.bar.stablebtn.texture_down)
						S.ShowStable()
					end
				end
			end
		end)
		Growler.bar.stablebtn.update = function()
			if not Growler.HasVisitedStables() then
				Growler.bar.stablebtn.texture:SetTexture(Growler.bar.stablebtn.texture_disabled)
			else
				if S.IsOpen then
					Growler.bar.stablebtn.texture:SetTexture(Growler.bar.stablebtn.texture_down)
				else
					Growler.bar.stablebtn.texture:SetTexture(Growler.bar.stablebtn.texture_up)
				end
			end
		end
		Growler.bar.stablebtn.setTooltipText = function()
			local tt = L["%sBrowse your stables."]:format(C.color.white)
			GameTooltip:SetText(tt)
		end
		Growler.bar.stablebtn.updateTooltip = function()
			if Growler.bar.stablebtn.mouseover then
				Growler.bar.stablebtn.setTooltipText()
			end
		end
		Growler.bar.stablebtn:SetScript("OnEnter",function(self)
			Growler.bar.stablebtn.mouseover = true
			GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
			Growler.bar.stablebtn.setTooltipText()
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("BOTTOMLEFT",Growler.bar.stablebtn,"TOPLEFT",0,0)
			GameTooltip:Show()
			Growler.bar.stablebtn.updateTooltip()
		end)
		Growler.bar.stablebtn:SetScript("OnLeave",function(self)
			Growler.bar.stablebtn.mouseover = false
			GameTooltip:ClearAllPoints()
			GameTooltip:Hide()
		end)
		
		
		Growler.bar.petslotbtn = CreateFrame("BUTTON","GrowlerBar_PetSlotButton",Growler.bar)
		Growler.bar.petslotbtn.pos = 3
		Growler.bar.petslotbtn.size = barbtnSize
		Growler.bar.petslotbtn.active = false
		Growler.bar.petslotbtn:SetSize(Growler.bar.petslotbtn.size,btnSize)
		Growler.bar.petslotbtn:SetPoint("RIGHT",Growler.bar,"RIGHT",((-Growler.bar.settingsbtn.size) - btnSpacing) * Growler.bar.petslotbtn.pos,0 * scale)
		Growler.bar.petslotbtn.texture_up = "Interface\\BUTTONS\\UI-MicroButton-Bug-Up"
		Growler.bar.petslotbtn.texture_down = "Interface\\BUTTONS\\UI-MicroButton-Bug-Down"
		Growler.bar.petslotbtn.texture_disabled = ""
		l.SetTexture(Growler.bar.petslotbtn.texture_disabled,Growler.bar.petslotbtn)
		Growler.bar.petslotbtn.texture:SetPoint("TOPLEFT",Growler.bar.petslotbtn,"TOPLEFT",0 * scale,20 * scale)
		Growler.bar.petslotbtn.texture:SetPoint("BOTTOMRIGHT",Growler.bar.petslotbtn,"BOTTOMRIGHT",0 * scale,(-2) * scale)
		Growler.bar.petslotbtn:RegisterForDrag("LeftButton")
		Growler.bar.petslotbtn:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
		Growler.bar.petslotbtn:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
		Growler.bar.petslotbtn:RegisterForClicks("LeftButtonUp")
		Growler.bar.petslotbtn:SetScript("OnMouseUp",function(self,button)
			if button == "LeftButton" then
				local activeslot = Growler.GetActivePetSlot()
				if activeslot ~= GrowlerDB.options.activepetslot then
					Growler.SetActivePetSlot()
				end
			end
		end)
		Growler.bar.petslotbtn.update = function()
			if InCombatLockdown() then
				Growler.bar.petslotbtn.texture:SetDesaturated(1)
			else
				local activeslot = Growler.GetActivePetSlot()
				if activeslot == 0 then
					Growler.bar.petslotbtn.texture:SetDesaturated(1)
					Growler.bar.petslotbtn.texture:SetTexture(Growler.bar.petslotbtn.texture_up)
					Growler.bar.petslotbtn.active = false
				elseif activeslot == GrowlerDB.options.activepetslot then
					Growler.bar.petslotbtn.texture:SetDesaturated(nil)
					Growler.bar.petslotbtn.texture:SetTexture(Growler.bar.petslotbtn.texture_down)
					Growler.bar.petslotbtn.active = true
				else
					Growler.bar.petslotbtn.texture:SetDesaturated(nil)
					Growler.bar.petslotbtn.texture:SetTexture(Growler.bar.petslotbtn.texture_up)
					Growler.bar.petslotbtn.active = false
				end
			end
		end
		Growler.bar.petslotbtn.setTooltipText = function()
			local tt = ""
			if Growler.bar.petslotbtn.active then
				tt = L["%sYour current pet slot(%s%s%s) is saved."]:format(C.color.white,C.color.green,tostring(Growler.GetActivePetSlot()),C.color.white)
			elseif Growler.GetActivePetSlot() ~= 0 then
				tt = L["%sLeft-click: %sSave this slot(%s%s%s) as your active pet slot."]:format("|r",C.color.white,C.color.green,tostring(Growler.GetActivePetSlot()),C.color.white)
			end
			GameTooltip:SetText(tt)
		end
		Growler.bar.petslotbtn.updateTooltip = function()
			if Growler.bar.petslotbtn.mouseover then
				Growler.bar.petslotbtn.setTooltipText()
			end
		end
		Growler.bar.petslotbtn:SetScript("OnEnter",function(self)
			Growler.bar.petslotbtn.mouseover = true
			GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
			Growler.bar.petslotbtn.setTooltipText()
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("BOTTOMLEFT",Growler.bar.petslotbtn,"TOPLEFT",0,0)
			GameTooltip:Show()
			Growler.bar.petslotbtn.updateTooltip()
		end)
		Growler.bar.petslotbtn:SetScript("OnLeave",function(self)
			Growler.bar.petslotbtn.mouseover = false
			GameTooltip:ClearAllPoints()
			GameTooltip:Hide()
		end)
		
		
		Growler.bar.update = function()
			Growler.bar.resetbtn.update()
			Growler.bar.settingsbtn.update()
			Growler.bar.lockbtn.update()
			Growler.bar.petslotbtn.update()
			Growler.bar.stablebtn.update()
			for slot=1,5 do
				Growler.petbtn[slot].update()
			end
			for spec=1,3 do
				Growler.specbtn[spec].update()
			end
			Growler.bar.PortraitFrame.update()
		end
				
		Growler.bar.SavePos = function()
			local point, relativeTo, relativePoint, xOfs, yOfs = Growler.bar:GetPoint()
			GrowlerDB.options.point, GrowlerDB.options.relativeto, GrowlerDB.options.relativepoint, GrowlerDB.options.pos_x, GrowlerDB.options.pos_y = point, relativeto, relativePoint, xOfs, yOfs
			point, relativeTo, relativePoint, xOfs, yOfs = GrowlerSI_PosFrame:GetPoint()
			GrowlerDB.options.si_point, GrowlerDB.options.si_relativeto, GrowlerDB.options.si_relativepoint, GrowlerDB.options.si_pos_x, GrowlerDB.options.si_pos_y = point, relativeto, relativePoint, xOfs, yOfs
			Growler.bar.ResetPos()
		end
		
		Growler.bar.ResetPos = function(force)
			if not InCombatLockdown() then
				if force then
					scale = Growler.DefaultDB.options.barsize
					siscale = Growler.DefaultDB.options.sisize
					GrowlerDB.options.barsize = scale
					GrowlerDB.options.sisize = siscale
				else
					scale = GrowlerDB.options.barsize
					siscale = GrowlerDB.options.sisize
				end
				btnPadding = C.btnPadding
				btnSpacing = C.btnSpacing
				barbtnSize = C.barbtnSize * scale
				btnSize = C.btnSize * scale
				petbtnSize = C.petbtnSize * scale
				petbtnOffset = C.petbtnOffset * scale
				siSize = C.siSize * siscale
				portraitSize = C.portraitSize * scale
				healthbarSize = portraitSize / 2.2
				if GrowlerDB.options.showportrait then
					stancebtnSize = ((btnSize + healthbarSize - 4 + (2 * btnPadding) - (2 * petbtnOffset)) / 3)
				else
					stancebtnSize = ((btnSize + (2 * btnPadding) - (2 * petbtnOffset)) / 3)
				end
				
				local spec = GetSpecialization()
				local _, _, _, selected, _, _, _, _  = GetTalentInfo(1, 1, 1)
				local hasLonewolf = (spec == 2 and selected)			
				local numbtns = 0
				
				local point, relativeto, relativepoint, pos_x, pos_y = C.Options.main.point.default, C.Options.main.relativeto.default, C.Options.main.relativepoint.default, C.Options.main.pos_x.default, C.Options.main.pos_y.default
				local si_point, si_relativeto, si_relativepoint, si_pos_x, si_pos_y = C.Options.main.si_point.default, C.Options.main.si_relativeto.default, C.Options.main.si_relativepoint.default, C.Options.main.si_pos_x.default, C.Options.main.si_pos_y.default
				if GrowlerDB.options.savepos and force ~= true then
					point, relativeto, relativepoint, pos_x, pos_y = GrowlerDB.options.point, GrowlerDB.options.relativeto, GrowlerDB.options.relativepoint, GrowlerDB.options.pos_x, GrowlerDB.options.pos_y
					si_point, si_relativeto, si_relativepoint, si_pos_x, si_pos_y = GrowlerDB.options.si_point, GrowlerDB.options.si_relativeto, GrowlerDB.options.si_relativepoint, GrowlerDB.options.si_pos_x, GrowlerDB.options.si_pos_y
				end
				
				if GrowlerDB.options.enable and GrowlerDB.options.enable_attack then
					numbtns = numbtns + 1
					Growler.btn.attack:Show()
					Growler.btn.attack.barpos = numbtns
				else
					Growler.btn.attack:Hide()
				end
				
				if GrowlerDB.options.enable and GrowlerDB.options.enable_command then
					numbtns = numbtns + 1
					Growler.btn.command:Show()
					Growler.btn.command.barpos = numbtns
				else
					Growler.btn.command:Hide()
				end
				
				if GrowlerDB.options.enable and (GrowlerDB.options.enable_hotp or GrowlerDB.options.enable_laststand) then
					numbtns = numbtns + 1
					Growler.btn.talent:Show()
					Growler.btn.talent.barpos = numbtns
				else
					Growler.btn.talent:Hide()
				end
				
				if GrowlerDB.options.enable and GrowlerDB.options.enable_special then
					numbtns = numbtns + 1
					Growler.btn.special:Show()
					Growler.btn.special.barpos = numbtns
				else
					Growler.btn.special:Hide()
				end
				
				if GrowlerDB.options.enable and GrowlerDB.options.enable_growl then
					numbtns = numbtns + 1
					Growler.btn.growl:Show()
					Growler.btn.growl.barpos = numbtns
				else
					Growler.btn.growl:Hide()
				end
				
				if GrowlerDB.options.enable and (GrowlerDB.options.enable_call or GrowlerDB.options.enable_revive or GrowlerDB.options.enable_mend) then
					numbtns = numbtns + 1
					Growler.btn.call:Show()
					Growler.btn.call.barpos = numbtns
				else
					Growler.btn.call:Hide()
				end
				
				if GrowlerDB.options.enable and l.PlayerHasSpell(GetSpellInfo(209997)) and (GrowlerDB.options.enable_playdead or GrowlerDB.options.enable_wakeup) then
					numbtns = numbtns + 1
					Growler.btn.playdead:Show()
					Growler.btn.playdead.barpos = numbtns
				else
					Growler.btn.playdead:Hide()
				end
				
				local barwidth = (((barbtnSize+btnSpacing)*3) + (btnPadding*2) + (numbtns * (btnSize+btnSpacing)))
				if GrowlerDB.options.showportrait then
					barwidth = ((portraitSize/2) + ((barbtnSize+btnSpacing)*3) + (btnPadding*2) + (numbtns * (btnSize+btnSpacing)))
				end
				local petbtnwidth = (3+btnSpacing + ((petbtnSize+btnSpacing) * 5))
				local specbtnwidth = (4+btnSpacing + ((petbtnSize+btnSpacing) * 3))
				if GrowlerDB.options.showpets and GrowlerDB.options.showspecs and GrowlerDB.options.petsbottom == GrowlerDB.options.specsbottom and barwidth < (petbtnwidth + specbtnwidth) then
					barwidth = (petbtnwidth + specbtnwidth)
				elseif GrowlerDB.options.showpets and barwidth < petbtnwidth then
					barwidth = petbtnwidth
				elseif GrowlerDB.options.showspecs and barwidth < specbtnwidth then
					barwidth = specbtnwidth
				end
				
				Growler.bar:SetMovable(true)
				Growler.bar:SetUserPlaced(false)
				Growler.bar:ClearAllPoints()
				Growler.bar:SetPoint(point,relativeto,relativepoint,pos_x,pos_y)
				Growler.bar:SetMovable(false)
				
				Growler.bar:SetMovable(true)
				Growler.bar:SetUserPlaced(false)
				GrowlerSI_PosFrame:ClearAllPoints()
				GrowlerSI_PosFrame:SetSize(siSize,siSize)
				GrowlerSI_PosFrame:SetPoint(si_point, si_relativeto, si_relativepoint, si_pos_x, si_pos_y)
				Growler.bar:SetMovable(false)
				
				Growler.bar.resetbtn.size = 15*scale
				Growler.bar.settingsbtn.size = 15*scale
				Growler.bar.resetbtn:SetSize(Growler.bar.resetbtn.size,Growler.bar.resetbtn.size)
				Growler.bar.settingsbtn:SetSize(Growler.bar.settingsbtn.size,Growler.bar.settingsbtn.size)
				Growler.bar.lockbtn:SetSize(barbtnSize,btnSize)
				Growler.bar.stablebtn:SetSize(barbtnSize,btnSize)
				Growler.bar.petslotbtn:SetSize(barbtnSize,btnSize)
				
				for k,btn in pairs(Growler.btn) do
					btn.hide()
					btn:ClearAllPoints()
					btn:SetSize(btnSize,btnSize)
					btn.text:SetTextHeight(16 * scale)
					local btnx = btnPadding + 2 + ((btnSize+btnSpacing) * (btn.barpos-1))
					if GrowlerDB.options.showportrait then
						btnx = (portraitSize/2) + btnPadding + 2 + ((btnSize+btnSpacing) * (btn.barpos-1))
					end
					btn:SetPoint("BOTTOMLEFT",Growler.bar,"BOTTOMLEFT",btnx,btnPadding)
					if btn.si then
						btn.si:ClearAllPoints()
						btn.si:SetSize(siSize,siSize)
						btn.si:SetPoint(si_point, si_relativeto, si_relativepoint, si_pos_x, si_pos_y)
					end
				end
				
				for spec=1,3 do
					local p = spec
					if spec == 3 then p = 1 elseif spec == 1 then p = 3 end
					local point, relativeto, relativepoint, y_offset = "BOTTOMLEFT", Growler.bar, "TOPRIGHT", -petbtnOffset
					if GrowlerDB.options.specsbottom then
						point, relativepoint, y_offset = "TOPLEFT", "BOTTOMRIGHT", -y_offset
						relativeto = GrowlerDB.options.showportrait and Growler.bar.PortraitFrame.healthbar or Growler.bar						
					end
					Growler.specbtn[spec]:ClearAllPoints()
					Growler.specbtn[spec]:SetSize(petbtnSize,petbtnSize)
					Growler.specbtn[spec]:SetPoint(point,relativeto,relativepoint, -2- ((petbtnSize+btnSpacing) * p), y_offset)
				end
				
				for slot=1,5 do
					local point, relativeto, relativepoint, y_offset = "BOTTOMRIGHT", Growler.bar, "TOPLEFT", -petbtnOffset
					if GrowlerDB.options.petsbottom then
						point, relativepoint, y_offset = "TOPRIGHT", "BOTTOMLEFT", -y_offset
						relativeto = GrowlerDB.options.showportrait and Growler.bar.PortraitFrame.healthbar or Growler.bar
					end
					Growler.petbtn[slot]:ClearAllPoints()
					Growler.petbtn[slot]:SetSize(petbtnSize,petbtnSize)
					Growler.petbtn[slot]:SetPoint(point,relativeto,relativepoint, 3+((petbtnSize+btnSpacing) * slot), y_offset)
				end
				
				for stance=1,3 do
					Growler.stancebtn[stance]:ClearAllPoints()
					Growler.stancebtn[stance]:SetSize(stancebtnSize,stancebtnSize)
					Growler.stancebtn[stance]:SetPoint("TOPLEFT",Growler.bar,"TOPRIGHT",0,((-4) * scale) - (stancebtnSize * (stance-1)))
				end
				
				Growler.bar:SetSize(barwidth, btnSize + (btnPadding*2))
				
				Growler.bar.stablebtn:SetPoint("RIGHT",Growler.bar,"RIGHT",((-Growler.bar.settingsbtn.size) - btnSpacing) * Growler.bar.stablebtn.pos,0 * scale)
				Growler.bar.stablebtn.texture:SetPoint("TOPLEFT",Growler.bar.stablebtn,"TOPLEFT",0 * scale,20 * scale)
				Growler.bar.stablebtn.texture:SetPoint("BOTTOMRIGHT",Growler.bar.stablebtn,"BOTTOMRIGHT",0 * scale,(-2) * scale)
				Growler.bar.lockbtn:SetPoint("RIGHT",Growler.bar,"RIGHT",((-Growler.bar.settingsbtn.size) - btnSpacing) * Growler.bar.lockbtn.pos,0 * scale)
				Growler.bar.lockbtn.texture:SetPoint("TOPLEFT",Growler.bar.lockbtn,"TOPLEFT",0 * scale,0 * scale)
				Growler.bar.lockbtn.texture:SetPoint("BOTTOMRIGHT",Growler.bar.lockbtn,"BOTTOMRIGHT",16 * scale,(-18) * scale)
				Growler.bar.petslotbtn:SetPoint("RIGHT",Growler.bar,"RIGHT",((-Growler.bar.settingsbtn.size) - btnSpacing) * Growler.bar.petslotbtn.pos,0 * scale)
				Growler.bar.petslotbtn.texture:SetPoint("TOPLEFT",Growler.bar.petslotbtn,"TOPLEFT",0 * scale,20 * scale)
				Growler.bar.petslotbtn.texture:SetPoint("BOTTOMRIGHT",Growler.bar.petslotbtn,"BOTTOMRIGHT",0 * scale,(-2) * scale)
				
				
				if GrowlerDB.options.showportrait then
					Growler.bar.PortraitFrame:Show()
					Growler.bar.PortraitFrame:ClearAllPoints()
					Growler.bar.PortraitFrame:SetSize(portraitSize,portraitSize)
					Growler.bar.PortraitFrame:SetPoint("RIGHT",Growler.bar,"LEFT",(portraitSize/2)-2*scale,-(portraitSize/4)+3*scale)
					Growler.bar.PortraitFrame.portraitframe:ClearAllPoints()
					Growler.bar.PortraitFrame.portraitframe:SetPoint("TOPLEFT",-portraitSize,portraitSize)
					Growler.bar.PortraitFrame.portraitframe:SetPoint("BOTTOMRIGHT",portraitSize,-portraitSize)
					Growler.bar.PortraitFrame.portrait:ClearAllPoints()
					Growler.bar.PortraitFrame.portrait:SetAllPoints(Growler.bar.PortraitFrame)
					SetPortraitTexture(Growler.bar.PortraitFrame.portrait,"pet")
					
					Growler.bar.PortraitFrame.healthbar:SetSize(barwidth,healthbarSize)
					Growler.bar.PortraitFrame.healthbar:SetPoint("TOPRIGHT",Growler.bar,"BOTTOMRIGHT",0,6)
					Growler.bar.PortraitFrame.healthbar.health:SetSize(barwidth-10,healthbarSize-10)
					Growler.bar.PortraitFrame.healthbar.health:SetPoint("TOPRIGHT",Growler.bar.PortraitFrame.healthbar,"TOPRIGHT",-5,-5)
				else
					Growler.bar.PortraitFrame:Hide()
				end
				
				if GrowlerDB.options.hidepetframe then
					PetFrame:SetScript("OnShow",function(self) if not InCombatLockdown() then self:Hide() end end)
					PetFrame:Hide()
				else
					PetFrame:SetScript("OnShow",nil)
					PetFrame:Show()
				end
				
				if GrowlerDB.options.hidedefaultbar then
					PetActionBarFrame:SetScript("OnShow",function(self) if not InCombatLockdown() then self:Hide() end end)
					-- PetActionBarFrame:SetScript("OnShow",function(self) self:Hide() end)
					PetActionBarFrame:Hide()
				else
					PetActionBarFrame:SetScript("OnShow",nil)
					PetActionBarFrame:Show()
				end
				
				if hasLonewolf then Growler.bar:Hide() else Growler.bar:Show() end
			else
				Growler.msg("Cannot do that in combat.",true)
			end
		end
	end
	
	Growler.CreateSpecButtons = function()
		for spec=1,3 do
			local icon = "Interface\\ICONS\\Ability_Druid_KingoftheJungle"
			if spec == 2 then
				icon = "Interface\\ICONS\\ABILITY_DRUID_DEMORALIZINGROAR"
			elseif spec == 3 then
				icon = "Interface\\ICONS\\Ability_EyeOfTheOwl"
			end
			local specbtn = CreateFrame("Button","GrowlerSpecBtn_" .. tostring(spec), Growler.bar, "SecureActionButtonTemplate")
			Growler.specbtn[spec] = specbtn
			specbtn:Show()
			specbtn:SetFrameLevel(0)
			specbtn.dpsremind = false
			specbtn.lastremind = 4.5
			specbtn.active = (GetSpecialization(false,true) == spec)
			specbtn:SetAttribute("type1","macro")
			specbtn:SetAttribute("macrotext1","/run SetSpecialization(" .. spec .. ",true)")
			specbtn:SetSize(petbtnSize,petbtnSize)
			specbtn:SetAlpha(1)
			specbtn:SetClampedToScreen(true)
			l.SetTexture(icon,specbtn)
			
			specbtn.setTooltipText = function()
				local tt = ""
				local talents = (spec==1 and "Ferocity" or (spec==2 and "Tenacity" or "Cunning"))
				if specbtn.active then
					tt = L["%sYour pet's currently active talents: %s"]:format(C.color.grey,talents)
				else
					tt = L["%sLeft-click: %sChange your pet's talents to %s."]:format("|r",C.color.white,C.color.green..talents..C.color.white)
				end
				GameTooltip:SetText(tt)
			end
			specbtn.updateTooltip = function()
				if specbtn.mouseover then
					specbtn.setTooltipText()
				end
			end
			specbtn:SetScript("OnEnter",function(self)
				specbtn.mouseover = true
				GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
				specbtn.setTooltipText()
				GameTooltip:ClearAllPoints()
				GameTooltip:SetPoint("BOTTOMLEFT",specbtn,"TOPLEFT",0,0)
				GameTooltip:Show()
				specbtn.updateTooltip()
			end)
			specbtn:SetScript("OnLeave",function(self)
				specbtn.mouseover = false
				GameTooltip:ClearAllPoints()
				GameTooltip:Hide()
			end)
			
			specbtn:RegisterForClicks("LeftButtonUp")
			specbtn:RegisterForDrag("LeftButton")
			specbtn:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
			specbtn:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
			specbtn:SetScript("OnMouseUp", function(self, button)
				if button == "RightButton" then
				elseif button == "MiddleButton" then
				elseif button == "LeftButton" then
				end
			end)
			specbtn.update = function()
				specbtn.lastremind = specbtn.lastremind + 0.5
				local hasPet, isHunterPet = HasPetUI()
				specbtn.active = (GetSpecialization(false,true) == spec)
				if specbtn.active then
					ActionButton_HideOverlayGlow(specbtn)
					specbtn.texture:SetDesaturated(nil)
					specbtn:RegisterForClicks()
				else
					if specbtn.lastremind > 4.5 then
						local inInstance, instanceType = IsInInstance()
						local _, instanceType, _, _, maxPlayers, _, _, _ = GetInstanceInfo()
						if GrowlerDB.options.dpsremind and spec == 1 and not InCombatLockdown() and IsInGroup() and maxPlayers>1 and InstanceType~="pvp" and hasPet and isHunterPet and not UnitUsingVehicle("player") and not IsMounted() and not IsFlying() and not UnitOnTaxi("player") then
							if not specbtn.dpsremind then
								if GrowlerDB.options.sound then PlaySoundKitID(31752) end
								ActionButton_ShowOverlayGlow(specbtn)
								specbtn.dpsremind = true
							end
							ActionButton_HideOverlayGlow(specbtn)
							ActionButton_ShowOverlayGlow(specbtn)
							if GrowlerDB.options.sound and GrowlerDB.options.reminder then PlaySoundKitID(38326) end
							specbtn.lastremind = 0
						else
							if specbtn.lastremind > 1 then
								ActionButton_HideOverlayGlow(specbtn)
								specbtn.lastremind = 0
							end
						end
					end
					specbtn.texture:SetDesaturated(1)
					specbtn:RegisterForClicks("LeftButtonUp")
				end
			end
		end
	end
	
	Growler.CreatePetButtons = function()
		for slot=1,5 do
			local petIcon, petName, petLevel, petType, petLoyalty = GetStablePetInfo(slot)
			local petbtn = CreateFrame("Button","GrowlerPetBtn_" .. tostring(slot), Growler.bar, "SecureActionButtonTemplate")
			Growler.petbtn[slot] = petbtn
			petbtn.slot = slot
			petbtn:Show()
			petbtn:SetFrameLevel(0)
			petbtn.active = (Growler.GetActivePetSlot() == petbtn.slot)
			petbtn.haspetinslot = (GetStablePetInfo(slot) ~= nil)
			petbtn.mouseover = false
			petbtn:SetAttribute("type1","macro")
			petbtn:SetAttribute("type2","macro")
			--petbtn:SetAttribute("type3","macro")
			petbtn:SetSize(petbtnSize,petbtnSize)
			petbtn:SetAlpha(1)
			petbtn:SetClampedToScreen(true)
			l.SetTexture(petIcon,petbtn)			
			
			petbtn.setTooltipText = function()
				local petIcon, petName, petLevel, petType, petLoyalty = GetStablePetInfo(petbtn.slot)
				petbtn.haspetinslot = (petIcon ~= nil)
				petbtn.active = (Growler.GetActivePetSlot() == petbtn.slot)
				local hasPet, isHunterPet = HasPetUI()
				local tt = ""
				if petbtn.active then
					tt = L["%s: %sDismiss your pet.%s"]:format(L["Right-click"],C.color.white,"\n\n")
					tt = tt.."|cff00ff00"..tostring(petName).." |cffffffff("..tostring(petLevel).." "..tostring(petType)..")"
				else
					if petbtn.haspetinslot then
						if hasPet and isHunterPet then
							tt = L["%s: %sDismiss your pet.%s"]:format(L["Click"],C.color.white,"\n\n")
							tt = tt.."|cff00ff00"..tostring(petName).." |cffffffff("..tostring(petLevel).." "..tostring(petType)..")"
						else
							tt = L["%s%s: %sSummon %s"]:format("|r",L["Left-click"],C.color.white,"|cff00ff00"..tostring(petName).." |cffffffff("..tostring(petLevel).." "..tostring(petType)..")")
						end
					else
						tt = L["%s: %sBeast Lore.%s"]:format(L["Right-click"],C.color.white,"\n")
						tt = tt .. L["%s%s: %sDismiss Pet / Tame Beast."]:format("\n|r",L["Left-click"],C.color.white)
					end
				end
				GameTooltip:SetText(tt)
			end
			petbtn.updateTooltip = function()
				if petbtn.mouseover then
					petbtn.setTooltipText()
				end
			end
			petbtn:SetScript("OnEnter",function(self)
				petbtn.mouseover = true
				GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
				petbtn.setTooltipText()
				GameTooltip:ClearAllPoints()
				GameTooltip:SetPoint("BOTTOMLEFT",petbtn,"TOPLEFT",0,0)
				GameTooltip:Show()
				petbtn.updateTooltip()
			end)
			petbtn:SetScript("OnLeave",function(self)
				petbtn.mouseover = false
				GameTooltip:ClearAllPoints()
				GameTooltip:Hide()
			end)
			
			petbtn:RegisterForClicks("LeftButtonUp","RightButtonUp")--,"MiddleButtonUp"
			petbtn:RegisterForDrag("LeftButton")
			petbtn:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
			petbtn:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
			petbtn:SetScript("OnMouseUp", function(self, button)
				if button == "RightButton" then
				elseif button == "MiddleButton" then
				elseif button == "LeftButton" then
				end
			end)
			petbtn.update = function()
				local spec = GetSpecialization()
				local _, _, _, selected, _, _, _, _  = GetTalentInfo(1, 1, 1)
				local hasLonewolf = (spec == 2 and selected)
				local petIcon, petName, petLevel, petType, petLoyalty = GetStablePetInfo(petbtn.slot)
				petbtn.haspetinslot = (petIcon ~= nil)
				petbtn.updateTooltip()
				
				if not InCombatLockdown() then
					petbtn.texture:SetTexture(petIcon)
					if petIcon == nil then
						petbtn.texture:SetTexture(l.GetSpellIcon(1515))
						petbtn:SetAttribute("macrotext1","/cast [nopet,btn:1,@target,harm,nodead,exists] "..GetSpellInfo(186265).."\n/cast [nopet,btn:1] "..GetSpellInfo(1515).."; [pet,btn:1] "..GetSpellInfo(2641))
						petbtn:SetAttribute("macrotext2","/cast [nopet,btn:2] "..GetSpellInfo(1462))
						petbtn.texture:SetVertexColor(1,1,1,1)
					elseif (not hasLonewolf and not l.pets.families[petType].exotic) or spec == 1 then
						--[@pet,nodead,pet:" .. tostring(petName) .. ",btn:1] Misdirection; 
						petbtn:SetAttribute("macrotext1","/cast [nopet,btn:1] "..GetSpellInfo(l.pets.callpet[petbtn.slot]) .. "; [nopet:" .. tostring(petName) .. ",btn:1] "..GetSpellInfo(2641))
						petbtn:SetAttribute("macrotext2","/cast [pet,btn:2] "..GetSpellInfo(2641))
						--petbtn:SetAttribute("macrotext3","/use [pet:" .. tostring(petName) .. ",btn:3] Essence Swapper")
						petbtn.texture:SetVertexColor(1,1,1,1)
					else
						petbtn:SetAttribute("macrotext1","/run Growler.msg('"..L["You cannot control exotic pets"]..".',true)")
						petbtn:SetAttribute("macrotext2","/run Growler.msg('"..L["You cannot control exotic pets"]..".',true)")
						petbtn.texture:SetVertexColor(1,0.2,0.2,1)
					end
				end
				
				if petbtn.haspetinslot then
					local hasPet, isHunterPet = HasPetUI()
					petbtn.active = (Growler.GetActivePetSlot() == petbtn.slot)
					if petbtn.active or not (hasPet and isHunterPet) then
						petbtn.texture:SetDesaturated(nil)
					else
						petbtn.texture:SetDesaturated(1)
					end
				else
					petbtn.texture:SetDesaturated(nil)
				end
			end
		end
	end
	
	
	Growler.CreateStanceButtons = function()
		for stance=1,3 do
			local iconName = (stance==1 and "ability_hunter_pet_assist") or (stance==2 and "ability_defend") or (stance==3 and "ability_seal")
			local stanceName = {"assist","defensive","passive"}
			local icon = "Interface\\ICONS\\"..iconName
			local stancebtn = CreateFrame("Button","GrowlerStanceBtn_" .. tostring(stance), Growler.bar, "SecureActionButtonTemplate")
			Growler.stancebtn[stance] = stancebtn
			stancebtn.name = stanceName[stance]
			stancebtn.stance = stance
			stancebtn:Show()
			stancebtn.isActiveButton = true
			stancebtn:SetFrameLevel(0)
			stancebtn.tooltip = L["Put your pet on %s."]:format(L[stanceName[stance]])
			stancebtn.mouseover = false
			stancebtn:SetAttribute("type1","macro")
			stancebtn:SetAttribute("macrotext1","/pet"..stanceName[stance])
			stancebtn:SetAlpha(1)
			stancebtn:SetClampedToScreen(true)
			stancebtn:RegisterForClicks("LeftButtonUp")
			stancebtn.lastupdate = 0
			stancebtn.snooze = -0.5
			stancebtn.active = false
			stancebtn.selected = false
			stancebtn.soundID = 12867
			stancebtn.originalIcon = icon
			stancebtn.currentIcon = nil
			l.SetTexture(icon,stancebtn)
			local text = stancebtn:CreateFontString("Growler_StanceBtn_"..tostring(stance).."_text", "OVERLAY", "DialogButtonNormalText")
			text:SetPoint("BOTTOM",stancebtn,"BOTTOM",0,2 * scale)
			text:SetTextHeight(16 * scale)
			stancebtn.text = text
			stancebtn.si = Growler.CreateSI("growl",icon)
			
			stancebtn.setTooltipText = function()
				local tt = ""
				if not stancebtn.isActiveButton then
					if stancebtn.snooze > 0 then
						local mins,secs = l.SecsToMins(stancebtn.snooze)
						tt = L["Snoozing: %s"]:format(C.color.green)
						if mins > 0 then
							tt = tt .. tostring(mins) .. "m "..L["and"].." "
						end
						tt = tt .. L["%ss %sleft.%s"]:format(tostring(secs),C.color.white,"\n\n")
						tt = tt .. L["%s%s: %sStop snoozing.%s"]:format("|r",L["Right-click"],C.color.white,"\n")
					end
					if stancebtn.active then
						tt = L["%s: %sSnooze for %s%s %sseconds.%s"]:format(L["Right-click"],C.color.white,C.color.green,tostring(math.floor(GrowlerDB.options.snooze)),C.color.white,"\n")
					end
					if stancebtn.active or stancebtn.snooze > 0 then
						tt = tt .. string.format("\n|r%s:\n",L["Left-click"])
					else
						tt = tt .. L["%sNo action required at the moment.%s"]:format(C.color.grey,"\n\n")
						tt = tt .. C.color.grey .. L["%s: %sSnooze for %s%s %sseconds.%s"]:format(L["Right-click"],"",C.color.green,tostring(math.floor(GrowlerDB.options.snooze)),C.color.grey,"\n")
						tt = tt .. C.color.grey .. L["Left-click"] .. ":\n"
					end
					
					tt = tt .. C.color.white .. stancebtn.tooltip
				else
					tt = stancebtn.tooltip
				end
				GameTooltip:SetText(tt)
			end
			stancebtn.updateTooltip = function()
				if stancebtn.mouseover then
					stancebtn.setTooltipText()
				end
			end
			stancebtn:SetScript("OnEnter",function(self)
				stancebtn.mouseover = true
				GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
				stancebtn.setTooltipText()
				GameTooltip:ClearAllPoints()
				GameTooltip:SetPoint("BOTTOMLEFT",stancebtn,"TOPLEFT",0,0)
				GameTooltip:Show()
				stancebtn.updateTooltip()
			end)
			stancebtn:SetScript("OnLeave",function(self)
				stancebtn.mouseover = false
				GameTooltip:ClearAllPoints()
				GameTooltip:Hide()
			end)
			stancebtn:RegisterForDrag("LeftButton")
			stancebtn:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
			stancebtn:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
			stancebtn:SetScript("OnMouseUp", function(self, button)
				if button == "RightButton" then
					if stancebtn.active then
						if stancebtn.snooze < 0 then
							stancebtn.snooze = GrowlerDB.options.snooze
							stancebtn.texture:SetVertexColor(0.3,0.9,1,0.5)
							stancebtn.texture:SetDesaturated(1)
						end
					else
						stancebtn:SetAlpha(1)
						stancebtn.snooze = 0
					end
				elseif button == "MiddleButton" then
				elseif button == "LeftButton" then
					if stancebtn.snooze > 0 then
						local soundID = 37656
						stancebtn.snooze = 0
						stancebtn.active = false
						ActionButton_HideOverlayGlow(stancebtn)
						stancebtn:SetAlpha(1)
						if GrowlerDB.options.sound then PlaySoundKitID(soundID) end
					else
						if not stancebtn.selected then
							for b,btn in pairs(Growler.stancebtn) do
								if stancebtn.stance==btn.stance then
									btn.selected = true
									btn.texture:SetDesaturated(nil)
									btn.texture:SetVertexColor(1,1,1,1)
								else
									btn.selected = false
									btn.texture:SetDesaturated(1)
									local desat = 0.4
									btn.texture:SetVertexColor(desat,desat,desat,1)
								end
							end
							if GrowlerDB.options.sound then PlaySoundKitID(11081) end
						end
					end
				end
			end)
			
			local stances = {"PET_MODE_ASSIST","PET_MODE_DEFENSIVE","PET_MODE_PASSIVE"}
		
			stancebtn.update = function()
				local hasPet, isHunterPet = HasPetUI()
				local petDead = UnitIsDead("pet")
				local desat = 0.4
				if hasPet and isHunterPet and not petDead then
					for i=1,NUM_PET_ACTION_SLOTS do
						local name, _, _, _, isActive = GetPetActionInfo(i)
						if name == stances[stancebtn.stance] then
							if not isActive and ((stancebtn.stance == 1 and GrowlerDB.options.enable_assist and TimeSinceLastPlayerAttack < 3 and InCombatLockdown()) or (stancebtn.stance == 2 and GrowlerDB.options.enable_defensive and TimeSinceLastPlayerAttack > 10 and not InCombatLockdown()) or (stancebtn.stance == 3 and GrowlerDB.options.enable_passive and TimeSinceLastPlayerAttack > 8 and TimeSinceLastPetAttack < 2)) then
								stancebtn.show()
							else
								stancebtn.hide()							
							end
							
							if stancebtn.snooze <= 0 then
								if isActive then
									stancebtn.texture:SetDesaturated(nil)
									stancebtn.texture:SetVertexColor(1,1,1,1)
								else
									stancebtn.selected = false
									stancebtn.texture:SetDesaturated(1)
									stancebtn.texture:SetVertexColor(desat,desat,desat,1)
								end
							end
							break
						end
					end
				else
					if stancebtn.snooze <= 0 then
						stancebtn.selected = false
						stancebtn.texture:SetDesaturated(1)
						stancebtn.texture:SetVertexColor(desat,desat,desat,1)
					end
					stancebtn.hide()
				end
			end
			
			stancebtn.show = function(icon,sound)
				Growler.ShowBtn(stancebtn,icon,sound)
			end
			stancebtn.hide = function(snooze)
				Growler.HideBtn(stancebtn,snooze)
			end
		end
	end
	
	Growler.CreatePortrait = function()
		Growler.bar.PortraitFrame = CreateFrame("Button","Growler_PortaitFrame",Growler.bar,"SecureActionButtonTemplate")
		local pf = Growler.bar.PortraitFrame
		pf:RegisterForClicks("AnyUp")
		pf:SetAttribute("type1","macro")
		pf:SetAttribute("type2","macro")
		pf:SetAttribute("type3","macro")
		pf:SetAttribute("macrotext1","/target pet")
		pf:SetAttribute("macrotext2","/cast "..GetSpellInfo(2641))
		pf:SetAttribute("macrotext3","/cast [@pet] "..GetSpellInfo(34477))
		pf:SetSize(portraitSize,portraitSize)
		pf:SetPoint("RIGHT",Growler.bar,"LEFT",0,-(portraitSize/4))
		
		pf:RegisterForDrag("LeftButton")
		pf:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
		pf:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
			
		pf.healthbar = CreateFrame("Button","Growler_HealthBar",pf)
		pf.healthbar:SetFrameLevel(0)
		pf.healthbar:RegisterForDrag("LeftButton")
		pf.healthbar:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
		pf.healthbar:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
		
		local backdrop = {
		  bgFile = "Interface\\FrameGeneral\\UI-Background-Rock",
		  edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
		  tile = false,
		  tileSize = 20,
		  edgeSize = 16,
		  insets = {left = 5,right = 5,top = 5,bottom = 5}
		}
		pf.healthbar:SetBackdrop(backdrop)
		pf.healthbar:SetBackdropColor(1,1,1,1)
		pf.healthbar:SetBackdropBorderColor(1,1,1,1)
		
		pf.healthbar.health = CreateFrame("StatusBar","Growler_HealthBar",pf.healthbar,"SecureUnitButtonTemplate")
		pf.healthbar.health:RegisterEvent("UNIT_HEALTH")
		pf.healthbar.health:RegisterEvent("UNIT_MAXHEALTH")
		pf.healthbar.health:RegisterEvent("UNIT_PET")
		pf.healthbar.health:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar")
				
		pf.healthbar.health:SetScript("OnEvent",function(self,event,...)
			if ((event == "UNIT_HEALTH" or event == "UNIT_MAXHEALTH") and ... == "pet") or event == "UNIT_PET" then
				local hasPet = (select(1,HasPetUI()) and select(2,HasPetUI()))
				if hasPet then
					pf.healthbar.health:SetMinMaxValues(0,UnitHealthMax("pet"))
					pf.healthbar.health:SetValue(UnitHealth("pet"))
				else
					pf.healthbar.health:SetMinMaxValues(0,1)
					pf.healthbar.health:SetValue(1)
				end
			end
		end)
		
		pf.portraitframe = pf:CreateTexture("Growler_PortaitFrame_Texture","BACKGROUND")
		pf.portraitframe:ClearAllPoints()
		pf.portraitframe:SetAllPoints(pf)
		pf.portraitframe:SetTexture("Interface\\COMMON\\portrait-ring-withbg")
		pf.portraitframe:SetPoint("TOPLEFT",-portraitSize,portraitSize)
		pf.portraitframe:SetPoint("BOTTOMRIGHT",portraitSize,-portraitSize)
		pf.portrait = pf:CreateTexture("Growler_Portait_Texture","ARTWORK")
		pf.portrait:ClearAllPoints()
		pf.portrait:SetAllPoints(pf)
		SetPortraitTexture(pf.portrait,"pet")
		
		pf.update = function()
			local hasPet = (select(1,HasPetUI()) and select(2,HasPetUI()))
			if hasPet then
				local hp = UnitHealth("pet")
				local maxhp = UnitHealthMax("pet")
				local r,g,b = 3-(hp*3/maxhp),(hp*1.15/maxhp)-0.15,0
				pf.healthbar.health:SetStatusBarColor(r,g,b)
				SetPortraitToTexture(Growler.bar.PortraitFrame.portrait,nil)
				SetPortraitTexture(Growler.bar.PortraitFrame.portrait,"pet")
			else
				SetPortraitToTexture(Growler.bar.PortraitFrame.portrait,"Interface\\CHARACTERFRAME\\TEMPORARYPORTRAIT-PET")
				pf.healthbar.health:SetStatusBarColor(1,1,1)
			end
			Growler.bar.PortraitFrame.portrait:ClearAllPoints()
			Growler.bar.PortraitFrame.portrait:SetAllPoints(Growler.bar.PortraitFrame)
		end
		pf.update()
	end
	
	Growler.ShowBtn = function(btn,icon,sound)
		btn.lastupdate = btn.lastupdate + 0.5
		if sound == nil then sound = btn.soundID end
		if icon ~= nil then
			if btn.currentIcon~=icon then
				btn.texture:SetTexture(icon)
				btn.si.texture:SetTexture(icon)
				btn.currentIcon = icon
				ActionButton_HideOverlayGlow(btn)
				ActionButton_ShowOverlayGlow(btn)
				if GrowlerDB.options.sound then PlaySoundKitID(sound) end
			end
		end
		if not btn.active then
			btn.si.show()
			btn.active = true
			ActionButton_ShowOverlayGlow(btn)
			if GrowlerDB.options.sound and (icon == nil or btn.currentIcon==icon) then PlaySoundKitID(sound) end
			btn:RegisterForClicks("LeftButtonUp")
			btn.lastupdate = 0
			btn:SetAlpha(1)
		else
			if btn.lastupdate > 4.5 then
				ActionButton_HideOverlayGlow(btn)
				ActionButton_ShowOverlayGlow(btn)
				btn.lastupdate = 0
				if GrowlerDB.options.sound and GrowlerDB.options.reminder then PlaySoundKitID(38326) end
			end
		end
	end
	Growler.HideBtn = function(btn,snooze)
		local soundID = 33338
		if btn.active then
			btn.si.hide()
			btn.active = false
			ActionButton_HideOverlayGlow(btn)
			if snooze then
				soundID = 31756
				btn:SetAlpha(0.8)
				btn:RegisterForClicks("LeftButtonUp")
			elseif not btn.isActiveButton then
				btn:SetAlpha(0.2)
				btn:RegisterForClicks()
			else
				btn:SetAlpha(1)
			end
			if GrowlerDB.options.sound then PlaySoundKitID(soundID) end
			btn.texture:SetTexture(btn.originalIcon)
			btn.si.texture:SetTexture(btn.originalIcon)
			btn.currentIcon = btn.originalIcon
		end
		if snooze and btn.snooze >= 0 then
			btn.text:SetText(tostring(l.SecsToMins(btn.snooze,true,true,true)))
		else
			btn.text:SetText()
		end
	end
	
	Growler.CreateButton = function(name,icon,priority,soundID,macro,tooltip,updateHandler,isActiveButton)
		if soundID == nil then soundID = 12867 end
		local btnName = "GrowlerBtn_" .. name
		local btn = CreateFrame("Button", btnName, Growler.bar, "SecureActionButtonTemplate")
		Growler.btn[name] = btn
		btn:Show()
		btn.macro = macro
		btn.barpos = l.tablesize(Growler.btn)
		btn.isActiveButton = isActiveButton
		if isActiveButton then
			btn:SetAlpha(1)
			btn:RegisterForClicks("AnyDown")
			btn.isReady = true
			btn:SetAttribute("type","macro")
			btn:SetAttribute("macrotext",macro)
		else
			btn:SetAlpha(0.2)
			btn:RegisterForClicks()
			btn:SetAttribute("type1","macro")
			btn:SetAttribute("macrotext1",macro)
		end
		btn.name = btnName
		btn.lastupdate = 0
		btn.snooze = -0.5
		btn.active = false
		btn.soundID = soundID
		btn.priority = priority
		btn.tooltip = tooltip
		btn.mouseover = false
		btn:SetSize(btnSize,btnSize)
		btn:SetClampedToScreen(true)
		btn.originalIcon = icon
		btn.currentIcon = icon
		l.SetTexture(icon,btn)
		local text = btn:CreateFontString(btnName .. "_text", "OVERLAY", "DialogButtonNormalText")
		text:SetPoint("BOTTOM",btn,"BOTTOM",0,2 * scale)
		text:SetTextHeight(16 * scale)
		btn.text = text
		btn.si = Growler.CreateSI("growl",icon)
		btn.update = function(hasPet, isHunterPet, petDead, petHealth, petHealthMax)
			if updateHandler then updateHandler(hasPet, isHunterPet, petDead, petHealth, petHealthMax) end
		end
		btn.setTooltipText = function()
			local tt = ""
			if not btn.isActiveButton then
				if btn.snooze > 0 then
					local mins,secs = l.SecsToMins(btn.snooze)
					tt = L["Snoozing: %s"]:format(C.color.green)
					if mins > 0 then
						tt = tt .. tostring(mins) .. "m "..L["and"].." "
					end
					tt = tt .. L["%ss %sleft.%s"]:format(tostring(secs),C.color.white,"\n\n")
					tt = tt .. L["%s%s: %sStop snoozing.%s"]:format("|r",L["Right-click"],C.color.white,"\n")
				end
				if btn.active then
					tt = L["%s: %sSnooze for %s%s %sseconds.%s"]:format(L["Right-click"],C.color.white,C.color.green,tostring(math.floor(GrowlerDB.options.snooze)),C.color.white,"\n")
				end
				if btn.active or btn.snooze > 0 then
					tt = tt .. string.format("\n|r%s:\n",L["Left-click"])
				else
					tt = tt .. L["%sNo action required at the moment.%s"]:format(C.color.grey,"\n\n")
					tt = tt .. C.color.grey .. L["%s: %sSnooze for %s%s %sseconds.%s"]:format(L["Right-click"],"",C.color.green,tostring(math.floor(GrowlerDB.options.snooze)),C.color.grey,"\n")
					tt = tt .. C.color.grey .. L["Left-click"] .. ":\n"
				end
				
				tt = tt .. C.color.white .. btn.tooltip
			else
				tt = btn.tooltip
			end
			GameTooltip:SetText(tt)
		end
		btn.updateTooltip = function()
			if btn.mouseover then
				btn.setTooltipText()
			end
		end
		btn:SetScript("OnEnter",function(self)
			btn.mouseover = true
			GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
			btn.setTooltipText()
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("BOTTOMLEFT",btn,"TOPLEFT",0,0)
			GameTooltip:Show()
			btn.updateTooltip()
		end)
		btn:SetScript("OnLeave",function(self)
			btn.mouseover = false
			GameTooltip:ClearAllPoints()
			GameTooltip:Hide()
		end)
		btn:RegisterForDrag("LeftButton")
		btn:SetScript("OnDragStart", function(self) Growler.bar.DragStart(self) end)
		btn:SetScript("OnDragStop", function(self) Growler.bar.DragStop(self) end)
		btn:SetScript("OnMouseUp", function(self, button)
			if button == "RightButton" then
				if not btn.isActiveButton then
					if btn.active then
						if btn.snooze < 0 then
							btn.snooze = GrowlerDB.options.snooze
							btn.texture:SetVertexColor(0.3,0.9,1,0.5)
							btn.texture:SetDesaturated(1)
						end
					else
						btn:SetAlpha(0.2)
						btn.snooze = 0
					end
				end
			elseif button == "MiddleButton" then
			elseif button == "LeftButton" then
			if not btn.isActiveButton then
				if btn.snooze > 0 then
					local soundID = 37656
					btn.snooze = 0
					btn.active = false
					ActionButton_HideOverlayGlow(btn)
					btn:SetAlpha(0.2)
					if GrowlerDB.options.sound then PlaySoundKitID(soundID) end
				end
				
				if self.active and self == GrowlerBtn_call then
					if GrowlerDB.options.activepetslot == 0 then
						Growler.msg(L["You do not have an active pet slot saved. Summon a pet and click the save button."],true)
					else
						Growler_PetCalled = 0
					end
				end
			end
			end
		end)
		btn.show = function(icon,sound)
			Growler.ShowBtn(btn,icon,sound)
		end
		btn.hide = function(snooze)
			Growler.HideBtn(btn,snooze)
		end
		
		return btn
	end
		
	Growler.CreateButtons = function()
		Growler.CreateButton("attack","Interface\\ICONS\\Ability_GhoulFrenzy",7,nil,
		"/petattack",
		L["%s%s: %sPet Attack"]:format("|r",L["Left-click"],C.color.white),
		function()
		end,true)
		
		local command = Growler.CreateButton("command","Interface\\ICONS\\Ability_Tracking",6,nil,
		"/petfollow [nomod,btn:1]\n/petmoveto [mod:ctrl,btn:1]\n/petstay [nomod,btn:2]",
		L["%s%s: %sPet Follow%s%s: %sPet Move-To%s%s: %sPet Stay"]:format("|r",L["Left-click"],C.color.white,"\n\n|r",L["Ctrl-Left-click"],C.color.white,"\n\n|r",L["Right-click"],C.color.white),
		function()
			local followormove = false
			for i=1,NUM_PET_ACTION_SLOTS do
				local name, _, _, _, isActive = GetPetActionInfo(i)
				if isActive then
					if name == "PET_ACTION_FOLLOW" then
						Growler.btn.command.texture:SetTexture("Interface\\ICONS\\Ability_Tracking")
						followormove = true
						break
					elseif name == "PET_ACTION_MOVE_TO" then
						Growler.btn.command.texture:SetTexture("Interface\\ICONS\\Ability_Hunter_Pet_GoTo")
						followormove = true
						break
					end
				end
			end
			
			if not followormove then
				Growler.btn.command.texture:SetTexture("Interface\\ICONS\\Spell_Nature_TimeStop")
			end
		end,true)
		
		
		Growler.CreateButton("talent",l.GetSpellIcon(88680),7,nil,
		"/cast "..GetSpellInfo(55709).."\n/cast "..GetSpellInfo(53478),
		"",function()
			local spec = GetSpecialization(false,true)
			local castSpell = ((spec==1 and GrowlerDB.options.enable_hotp) and 55709) or ((spec==2 and GrowlerDB.options.enable_laststand) and 53478) or nil			
			local hasPet,isHunterPet = HasPetUI()
			Growler.btn.talent.tooltip = ""
			if hasPet and isHunterPet and castSpell and (castSpell ~= 55709 or UnitIsDead("pet")) then
				Growler.btn.talent.texture:SetTexture(l.GetSpellIcon(castSpell))
				if l.SpellIsReady(GetSpellInfo(castSpell)) then
					Growler.btn.talent.tooltip = GetSpellBaseCooldown(castSpell) > 0 and "|r"..L["Cooldown"]..": |cff00ff00"..l.SecsToMins(GetSpellBaseCooldown(castSpell)/1000,true)..select(2,l.SecsToMins(GetSpellBaseCooldown(castSpell)/1000,true)).."\n\n" or ""
					Growler.btn.talent:SetAlpha(1)
					Growler.btn.talent.texture:SetDesaturated(nil)
					Growler.btn.talent.texture:SetVertexColor(1,1,1,1)
					Growler.btn.talent.text:SetText()
				else
					Growler.btn.talent:SetAlpha(0.8)
					Growler.btn.talent.texture:SetDesaturated(1)
					Growler.btn.talent.texture:SetVertexColor(0.5,0.5,0.5,1)
					Growler.btn.talent.text:SetText(tostring(l.SecsToMins(l.CooldownLeft(castSpell),true)))
					local cdleft = tostring(l.SecsToMins(l.CooldownLeft(castSpell)))
					Growler.btn.talent.tooltip = l.CooldownLeft(castSpell) > 0 and "|r"..L["Cooldown"]..": |cff00ff00"..l.SecsToMins(l.CooldownLeft(castSpell),true)..select(2,l.SecsToMins(l.CooldownLeft(castSpell),true)).."\n\n" or ""
				end
				Growler.btn.talent.tooltip = Growler.btn.talent.tooltip.."|r"..GetSpellInfo(castSpell).."\n|cffffffff"..GetSpellDescription(castSpell)
			else
				if hasPet and isHunterPet and castSpell and not UnitIsDead("pet") then
					Growler.btn.talent.texture:SetTexture(l.GetSpellIcon(castSpell))
					Growler.btn.talent.tooltip = Growler.btn.talent.tooltip..L["%sYour pet is not dead%s"]:format(C.color.grey,"\n\n")
					if l.SpellIsReady(GetSpellInfo(castSpell)) then
						Growler.btn.talent:SetAlpha(0.2)
						Growler.btn.talent.texture:SetDesaturated(1)
						Growler.btn.talent.texture:SetVertexColor(1,1,1,1)
						Growler.btn.talent.text:SetText()
						Growler.btn.talent.tooltip = GetSpellBaseCooldown(castSpell) > 0 and Growler.btn.talent.tooltip.."|r"..L["Cooldown"]..": |cff00ff00"..l.SecsToMins(GetSpellBaseCooldown(castSpell)/1000,true)..select(2,l.SecsToMins(GetSpellBaseCooldown(castSpell)/1000,true)).."\n\n" or ""
					else
						Growler.btn.talent.text:SetText(tostring(l.SecsToMins(l.CooldownLeft(castSpell),true)))
						Growler.btn.talent:SetAlpha(0.8)
						Growler.btn.talent.texture:SetDesaturated(1)
						Growler.btn.talent.texture:SetVertexColor(0.5,0.5,0.5,1)
						local cdleft = tostring(l.SecsToMins(l.CooldownLeft(castSpell)))
						Growler.btn.talent.tooltip = l.CooldownLeft(castSpell) > 0 and Growler.btn.talent.tooltip.."|r"..L["Cooldown"]..": |cff00ff00"..l.SecsToMins(l.CooldownLeft(castSpell),true)..select(2,l.SecsToMins(l.CooldownLeft(castSpell),true)).."\n\n" or ""
					end
					Growler.btn.talent.tooltip = Growler.btn.talent.tooltip.."|r"..GetSpellInfo(castSpell).."\n|cffffffff"..GetSpellDescription(castSpell)
				else
					Growler.btn.talent.tooltip = L["%sThere is no special ability for the current talent set"]:format(C.color.grey)
					Growler.btn.talent.texture:SetTexture(l.GetSpellIcon(88680))
					Growler.btn.talent:SetAlpha(0.2)
					Growler.btn.talent.texture:SetDesaturated(1)
					Growler.btn.talent.texture:SetVertexColor(1,1,1,1)
					Growler.btn.talent.text:SetText()
				end
			end
		end,true)
		
		local macro = ""
		for k,v in pairs(l.pets.specialAbilities) do
			macro = macro.."/cast "..GetSpellInfo(v).."\n"
		end		
		Growler.CreateButton("special",l.GetSpellIcon(88680),5,nil,
		macro,
		"",function()
			Growler.btn.special.lastupdate = Growler.btn.special.lastupdate + 0.5
			
			if Growler.btn.special.lastupdate > 0 then
				Growler.btn.special.lastupdate = 0
				
				local macro = ""
				local specialability = nil
				for k,v in pairs(l.pets.specialAbilities) do
					macro = macro.."/cast "..GetSpellInfo(v).."\n"
					if l.PlayerHasSpell(GetSpellInfo(v)) then
						specialability = v
					end
				end
				
				local hasPet,isHunterPet = HasPetUI()
				if hasPet and isHunterPet and specialability then
					Growler.btn.special.tooltip = ""
					Growler.btn.special.texture:SetTexture(l.GetSpellIcon(specialability))
					if l.SpellIsReady(GetSpellInfo(specialability)) then
						Growler.btn.special.tooltip = GetSpellBaseCooldown(specialability) > 0 and "|r"..L["Cooldown"]..": |cff00ff00"..l.SecsToMins(GetSpellBaseCooldown(specialability)/1000,true)..select(2,l.SecsToMins(GetSpellBaseCooldown(specialability)/1000,true)).."\n\n" or ""
						Growler.btn.special:SetAlpha(1)
						Growler.btn.special.texture:SetDesaturated(nil)
						Growler.btn.special.texture:SetVertexColor(1,1,1,1)
						Growler.btn.special.text:SetText()
					else
						Growler.btn.special:SetAlpha(0.8)
						Growler.btn.special.texture:SetDesaturated(1)
						Growler.btn.special.texture:SetVertexColor(0.5,0.5,0.5,1)
						local timeleft = l.SecsToMins(l.CooldownLeft(specialability)) > 0 and tostring(l.SecsToMins(l.CooldownLeft(specialability),true)) or tostring(select(2,l.SecsToMins(l.CooldownLeft(specialability),true)))
						Growler.btn.special.text:SetText(timeleft)
						local cdleft = tostring(l.SecsToMins(l.CooldownLeft(specialability)))
						Growler.btn.special.tooltip = GetSpellBaseCooldown(specialability) > 0 and "|r"..L["Cooldown"]..": |cff00ff00"..l.SecsToMins(l.CooldownLeft(specialability),true)..select(2,l.SecsToMins(l.CooldownLeft(specialability),true)).."\n\n" or ""
					end
					Growler.btn.special.tooltip = Growler.btn.special.tooltip.."|r"..GetSpellInfo(specialability).."\n|cffffffff"..GetSpellDescription(specialability)
				else
					Growler.btn.special.tooltip = L["%sYour pet does not have a special ability"]:format(C.color.grey)
					Growler.btn.special.texture:SetTexture(l.GetSpellIcon(88680))
					Growler.btn.special:SetAlpha(0.2)
					Growler.btn.special.texture:SetDesaturated(1)
					Growler.btn.special.texture:SetVertexColor(1,1,1,1)
					Growler.btn.special.text:SetText()
				end
			end
		end,true)
		
		local growlMacro = "/petautocaston "..GetSpellInfo(2649).."\n/stopmacro [talent:3,@boss1,nodead,exists]\n/petautocastoff [@party1,exists][@raid1,exists] "..GetSpellInfo(2649)
		Growler.CreateButton("growl",l.GetSpellIcon(2649),1,nil,
		growlMacro,
		L["Turns off Growl autocast when in a group.%sTurns on Growl autocast when alone.%sAlso turn on autocast for all pet autocast abilities enabled through the options menu"]:format("\n","\n\n"),
		function(hasPet, isHunterPet, petDead)
			local autocastable, autostate = GetSpellAutocast(select(1,GetSpellInfo(2649)))
			local inInstance, instanceType = IsInInstance()
			local _, instanceType, _, _, maxPlayers, _, _, _ = GetInstanceInfo()
			if Growler.PetAvailable then
				if l.PlayerHasSpell(GetSpellInfo(2649)) and GrowlerDB.options.enable_growl and instanceType ~= "pvp" and ((IsInGroup() and autostate==true and maxPlayers > 1) or (((not IsInGroup() or maxPlayers==1) and autostate==false) and GrowlerDB.options.nonparty)) then
					Growler.btn.growl.show(l.GetSpellIcon(2649))
				else
					local autocastiswrong = nil
					if GrowlerDB.options.enable_autocast then
						for k,v in pairs(l.pets.autocastAbilities) do
							for i,id in pairs(v) do
								if l.PlayerHasSpell(select(1,GetSpellInfo(id))) then
									if (GrowlerDB.options["enable_"..k] and not select(2,GetSpellAutocast(select(1,GetSpellInfo(id))))) or (not GrowlerDB.options["enable_"..k] and select(2,GetSpellAutocast(select(1,GetSpellInfo(id))))) then
										autocastiswrong = id
										break
									end
								end
							end
						end
					end
					if autocastiswrong ~= nil then
						Growler.btn.growl.show(l.GetSpellIcon(autocastiswrong))
					else
						Growler.btn.growl.hide()
					end
				end
			else
				Growler.btn.growl.hide()
			end
		end)
		Growler.btn.growl.updatemacro = function()
			if not InCombatLockdown() then
				local growlMacro = ""
				if GrowlerDB.options.enable_autocast then
					for k,v in pairs(l.pets.autocastAbilities) do
						local toggle = "off"
						if GrowlerDB.options["enable_"..k] then
							toggle = "on"
						end
						for i,id in pairs(v) do
							growlMacro = growlMacro .. "/petautocast"..toggle.." "..GetSpellInfo(id) .. "\n"
						end
					end
				end
				growlMacro = growlMacro .. "/petautocaston "..GetSpellInfo(2649).."\n"
				growlMacro = growlMacro .. "/petautocastoff [@party1,exists][@raid1,exists] "..GetSpellInfo(2649)
				
				Growler.btn.growl:SetAttribute("macrotext1",growlMacro)
			else
				Growler.msg(L["Cannot update this setting while in combat."],true)
			end
		end
		Growler.btn.growl.updatemacro()
		
		local stdicon = (GrowlerDB.options.enable_call and 23498) or (GrowlerDB.options.enable_revive and 982) or (GrowlerDB.options.enable_mend and 136) or 23498
		Growler.CreateButton("call",l.GetSpellIcon(stdicon),5,nil,
		"",
		L["Calls your saved pet if you have no pet.%sRevives your pet if it's dead.%sHeals your pet if it's low on health."]:format("\n","\n"),
		function(hasPet, isHunterPet, petDead, petHealth, petHealthMax)
			if GrowlerDB.options.enable_call and GrowlerDB.options.activepetslot>0 and not hasPet and WaitForPet > 1.5 and (not IsFalling() or IsJumping) and not IsMounted() and Growler_PetCalled > 3 then
				Growler.btn.call.show(l.GetSpellIcon(23498),31754)
			else
				if l.PlayerHasSpell(select(1,GetSpellInfo(982))) and GrowlerDB.options.enable_revive and (petDead or (not hasPet and not IsMounted() and GrowlerDB.options.enable_call and Growler_PetCalled <= 3 and Growler_PetCalled > 1)) then
					 -- for heart of the phoenix
					 -- and (GetSpecialization(false,true)~=1 or (not l.SpellIsReady(GetSpellInfo(55709))))
					Growler.btn.call.show(l.GetSpellIcon(982),38323)
				elseif l.PlayerHasSpell(select(1,GetSpellInfo(136))) and GrowlerDB.options.enable_mend and hasPet and isHunterPet and ((petHealth/petHealthMax) < GrowlerDB.options.mend_threshhold) and l.SpellIsReady("Mend Pet") then
					Growler.btn.call.show(l.GetSpellIcon(136),37656)
				else
					Growler.btn.call.hide()
				end
			end
		end)
		Growler.btn.call.updatemacro = function()
			if not InCombatLockdown() then
				Growler.btn.call.originalIcon = (GrowlerDB.options.enable_call and l.GetSpellIcon(23498)) or (GrowlerDB.options.enable_revive and "Interface\\ICONS\\ability_hunter_beastsoothe") or (GrowlerDB.options.enable_mend and l.GetSpellIcon(136)) or l.GetSpellIcon(23498)
				Growler.btn.call.texture:SetTexture(Growler.btn.call.originalIcon)
				local callpet = ""
				local revivepet = ""
				local mendpet = ""
				if GrowlerDB.options.enable_call and GrowlerDB.options.activepetslot>0 then
					callpet = "\n/castsequence [nopet] reset=3 "..GetSpellInfo(l.pets.callpet[GrowlerDB.options.activepetslot])..", "..GetSpellInfo(982)
				end
				if GrowlerDB.options.enable_revive then
					revivepet = "\n/cast [@pet,dead,exists] "..GetSpellInfo(982)
				end
				if GrowlerDB.options.enable_mend then
					mendpet = "\n/cast [@pet,nodead,exists] "..GetSpellInfo(136)
				end
				
				Growler.btn.call:SetAttribute("macrotext1",callpet .. revivepet .. mendpet)
			else
				Growler.msg(L["Cannot update this setting while in combat."],true)
			end
		end
		Growler.btn.call.updatemacro()
				
		Growler.CreateButton("playdead",l.GetSpellIcon(209997),4,34153,
		"/cast "..GetSpellInfo(209997),
		L["Your pet will Play Dead - available when you are feigning death.%sYour pet will Wake Up - available when it's playing dead and you are not."]:format("\n"),
		function(hasPet, isHunterPet, petDead)
			local isFeigning = l.UnitHasBuff("player",select(1,GetSpellInfo(5384)))
			local playsDead = l.UnitHasBuff("pet",select(1,GetSpellInfo(209997)))
			if l.PlayerHasSpell(select(1,GetSpellInfo(209997))) and GrowlerDB.options.enable_playdead and hasPet and isHunterPet and not petDead and isFeigning and not playsDead and l.SpellIsReady(209997) then
				Growler.btn.playdead.show(l.GetSpellIcon(209997),34153)
			elseif l.PlayerHasSpell(select(1,GetSpellInfo(209997))) and GrowlerDB.options.enable_wakeup and hasPet and isHunterPet and not petDead and not isFeigning and playsDead and l.SpellIsReady(210000) then
				Growler.btn.playdead.show(l.GetSpellIcon(210000),34152)
			else
				if not TipPlayDead and UnitLevel("player") == 110 and not l.PlayerHasSpell(select(1,GetSpellInfo(209997))) then
					Growler.msg("You have not learned the spell 'Play Dead'. Buy the book from your class hall for extra functionality.")
					TipPlayDead = true
				end
				Growler.btn.playdead.hide()
			end
		end)
	end
	
	local UpdateBtn = function(btn,hasPet, isHunterPet, petDead, petHealth, petHealthMax)
		if btn.snooze >= 0 then
			btn.snooze = btn.snooze - 0.5
			btn.hide(true)
		elseif btn.snooze <= 0 then
			if not btn.isActiveButton then
				btn.texture:SetVertexColor(1,1,1,1)
				btn.texture:SetDesaturated(nil)
			end
			if btn.isActiveButton or not (GrowlerDB.options.disablewhenresting and IsResting() and not InCombatLockdown()) then
				btn.update(hasPet, isHunterPet, petDead, petHealth, petHealthMax)
			else
				btn.hide()
			end
		end
		btn.updateTooltip()
	end	
	
	Growler.Update = function(self, elapsed)
		local spec = GetSpecialization()
		local _, _, _, selected, _, _, _, _  = GetTalentInfo(1, 1, 1)
		local hasLonewolf = (spec == 2 and selected)
		
		Growler.LastMessageCheck = Growler.LastMessageCheck + elapsed
		Growler.TalentChangeLastMsg = Growler.TalentChangeLastMsg + elapsed
		Growler.TalentChangeLastUpdate = Growler.TalentChangeLastUpdate + elapsed
		TimeSinceLastPetAttack = TimeSinceLastPetAttack + elapsed
		TimeSinceLastPlayerAttack = TimeSinceLastPlayerAttack + elapsed
		TimeSinceLastScreenIconUpdate = TimeSinceLastScreenIconUpdate + elapsed
		Growler_PetCalled = Growler_PetCalled + elapsed
		TimeSinceLastUpdate = TimeSinceLastUpdate + elapsed
		
		if isHunter and IsFalling() then
			TimeFalling = TimeFalling + elapsed
			if TimeFalling > 0 and TimeFalling < 2 then
				IsJumping = true
			else
				IsJumping = false
			end
		else
			TimeFalling = 0
			IsJumping = false
		end
		
		if TimeSinceLastUpdate > 0.5 then
			TimeSinceLastUpdate = 0
			
			if GrowlerDB.options.notifygroup and Growler.LastMessageCheck > 2 then
				Growler.LastMessageCheck = 0
				if Growler.GrowlWasCast and not Growler.GrowlMessageSent then
					Growler.GrowlMessageSent = true
					local myName = UnitName("player")
					local PeopleWhoSpottedGrowl = {myName}
					local i = 2
					for k,v in pairs(Growler.RecievedMessages) do
						if l.split(v,";")[1] == "growl" then PeopleWhoSpottedGrowl[i] = k end
						i = i + 1
					end
					table.sort(PeopleWhoSpottedGrowl)
					if myName == PeopleWhoSpottedGrowl[1] then
						local groupType = "PARTY"
						if IsInRaid() then groupType = "RAID" end
						SendChatMessage("{{Growler "..L["alert :: A hunter or warlock's pet is casting their taunt ability (growl / suffering / seethe) -- Download the addon to help manage your pets."].."}}",groupType)
					end		
				end
			end
			
			if isHunter then
				local hasPet, isHunterPet = HasPetUI()
				
				
				IsJumpingNotFalling = (IsJumping or not IsFalling())
				IsFallingNotJumping = (IsFalling() and not IsJumping)
				Growler.PetAvailable = (hasPet and isHunterPet and IsJumpingNotFalling and not IsMounted())
				
				if IsFallingNotJumping or IsFlying() or IsMounted() or UnitOnTaxi("player") or UnitUsingVehicle("player") then
					WaitForPet = 0
				elseif WaitForPet < 3 then
					WaitForPet = WaitForPet + 0.5
				else
					WaitForPet = 3
				end
				
				local isBusy = IsFlying() or UnitOnTaxi("player") or C_PetBattles.IsInBattle() or UnitUsingVehicle("player") or UnitIsDeadOrGhost("player") or HasVehicleActionBar() or HasTempShapeshiftActionBar() or HasBonusActionBar() or HasOverrideActionBar() or IsPossessBarVisible()
				-- local isBusy = IsFlying() or UnitOnTaxi("player") or C_PetBattles.IsInBattle() or UnitUsingVehicle("player") or UnitIsDeadOrGhost("player") or HasVehicleActionBar() -- or HasTempShapeshiftActionBar() or HasBonusActionBar() or HasOverrideActionBar() or IsPossessBarVisible() -- or HasExtraActionBar()
								
				if Growler.TalentOrButtonChanged or Growler.RefreshKeybinds then
					if not InCombatLockdown() then
						Growler.SetBindings()
						Growler.TalentChangeMsgShown = true
						Growler.TalentChangeLastMsg = 0
					elseif not Growler.TalentChangeMsgShown and Growler.TalentOrButtonChanged and Growler.TalentChangeLastMsg > 10 then
						Growler.msg(L["You changed a talent or an action button while in combat, and it will not be updated before you leave combat. This is an effect of the autocast feature, and can be prevented by turning it off."],true)
						Growler.TalentChangeMsgShown = true
						Growler.TalentChangeLastMsg = 0
					end
				end
				
				if not GrowlerDB.options.hidebar and not hasLonewolf and not isBusy then
					if not Growler.bar:IsShown() and not InCombatLockdown() then
						Growler.bar.ResetPos()
						Growler.bar:Show()
					end
					if GrowlerDB.options.enable and Growler.EnteredWorld then
						local petHealth = UnitHealth("pet")
						local petHealthMax = UnitHealthMax("pet")
						local petDead = UnitIsDead("pet")						
						
						for k,btn in pairs(Growler.btn) do
							UpdateBtn(btn,hasPet, isHunterPet, petDead, petHealth, petHealthMax)
						end
						for k,btn in pairs(Growler.stancebtn) do
							UpdateBtn(btn,hasPet, isHunterPet, petDead, petHealth, petHealthMax)
						end
						Growler.UpdateSI()
					end
					Growler.bar.update()
				elseif Growler.bar:IsShown() and (hasLoneWolf or GrowlerDB.options.hidebar or isBusy) then
					if not InCombatLockdown() then
						Growler.bar.ResetPos()
						Growler.bar:Hide()
					end
				end
			end
		end
	end
	
	local applicablespells = {
		feigndeath = 5384,
		
		cobrashot = 193455,
		killcommand = 34026,
		multishot = 2643,
		direbeast = 120679,
		concussiveshot = 5116,
		barrage = 120360,
		direfrenzy = 217200,
		chimaerashot = 53209,
		amurderofcrows_bm = 131894,
		amurderofcrows_mm = 206505,
		blackarrow = 194599,
		explosiveshot = 212431,
		sidewinders = 214579,
		piercingshot = 198670,
		aimedshot = 19434,
		arcaneshot = 185358,
		burstingshot = 186387,
		countershot = 147362,
		markedshot1 = 185901,
		markedshot2 = 231551,
		carve = 187708,
		flankingstrike = 202800,
		harpoon1 = 190925,
		harpoon2 = 231550,
		hatchettoss = 193265,
		lacerate = 185855,
		mogoosebite = 190928,
		muzzle = 187707,
		raptorstrike = 186270,
		wingclip = 195645,
		throwingaxes = 200163,
		butchery = 212436,
		spittingcobra = 194407,
		
		dragonsfiregrenade = 194855,
		stickybomb = 191241,
		
		-- explosivetrap = 191433,
		-- tartrap = 187698,
		-- caltrops = 194277,
		-- steeltrap = 162488,
		-- rangersnet = 200108,
	}
	
	Growler.SetBindings = function()
		if not InCombatLockdown() and isHunter then
			ClearOverrideBindings(GrowlerBar)
			local spec = GetSpecialization()
			local _, _, _, selected, _, _, _, _  = GetTalentInfo(1, 1, 1)
			local hasLonewolf = (spec == 2 and selected)
			if not hasLonewolf then--and (GrowlerDB.options.auto_playdead or GrowlerDB.options.auto_wakeup or GrowlerDB.options.auto_call or GrowlerDB.options.auto_growl or GrowlerDB.options.auto_attack or GrowlerDB.options.auto_revive) then
				for i=1,GetNumBindings() do
					local command, x, bind1, bind2 = GetBinding(i)
					if command then
						local actionbtn = string.match(command,"^ACTIONBUTTON(%d+)")
						-- local bonusactionbutton = string.match(command,"BONUSACTIONBUTTON(%d+)")
						local multiactionbar, multiactionbtn = string.match(command,"^MULTIACTIONBAR(%d+)BUTTON(%d+)")
						if actionbtn or multiactionbar then--or bonusactionbutton then
							local id
							if actionbtn then
								id = tonumber(actionbtn)
							elseif bonusactionbutton then
								id = tonumber(bonusactionbutton) + 12
							elseif multiactionbar=="3" then
								id = tonumber(multiactionbtn) + 24
							elseif multiactionbar=="4" then
								id = tonumber(multiactionbtn) + 36
							elseif multiactionbar=="2" then
								id = tonumber(multiactionbtn) + 48
							elseif multiactionbar=="1" then
								id = tonumber(multiactionbtn) + 60
							end
							
							-- Growler.SetBinding(id)
							repeat
								local actiontype, actionid = GetActionInfo(id)
								if (bind1 or bind2) and (actiontype == "spell" or actiontype == "macro") then
									local macrobody = nil
									local spellmatch = nil
									if actiontype == "spell" then
										for n,osid in pairs(applicablespells) do
											if tonumber(actionid)==osid then
												spellmatch = osid
												macrobody = "/cast "..select(1,GetSpellInfo(tonumber(actionid)))
												break
											end
										end
										if not spellmatch then do break end end
									elseif actiontype == "macro" then
										local tempbody = select(3,GetMacroInfo(actionid))
										for n,osid in pairs(applicablespells) do
											local osname = l.replace(select(1,GetSpellInfo(osid)),"-","%-")
											local match = tempbody~=nil and string.match(tempbody,".*("..select(1,GetSpellInfo(osid))..").*") or nil
											if match then
												spellmatch = osid
												macrobody = tempbody
												break
											elseif IsAddOnLoaded("M6") then
												-- check for M6
												local M6slot = tempbody~=nil and string.match(tempbody,".*#m6.*M6Prime%s(s[01][0-9a-f])") or nil
												if M6slot then
													local M6player = M6DB.profiles[GetRealmName()][GetUnitName("player")]
													local M6n
													for k=1,#M6player do
														if M6player[k] and l.tablesize(M6player[k]["slots"]) > 0 then
															M6n = k
															break
														end												
													end
													if M6n then
														local M6id = M6player[M6n]["slots"][M6slot]
														
														local match = M6DB.actions[M6id][2]~= nil and string.match(M6DB.actions[M6id][2],".*("..osname..").*") or nil
														if match then
															spellmatch = osid
															macrobody = M6DB.actions[M6id][2]
															break
														end
													end
												end
											end
										end
										if not spellmatch then do break end end
									end
									
									if macrobody and spellmatch then
									
										local extramacro = "#growler\n"
										for i=1,12 do
											if tonumber(bind1) == i or tonumber(bind2) == i then
												extramacro = extramacro .. "/click [vehicleui][overridebar][possessbar,@vehicle,exists] OverrideActionBarButton"..tostring(i).."; [possessbar][bonusbar][shapeshift] ActionButton"..tostring(i).."\n/stopmacro [vehicleui][possessbar][overridebar][shapeshift][bonusbar]\n"
												break
											end
										end
										extramacro = extramacro .. "/console Sound_EnableAllSound 0\n"
										if spellmatch == 5384 then
											extramacro = GrowlerDB.options.auto_playdead and extramacro .. "/cast ".. GetSpellInfo(209997) .. "\n" or extramacro
										elseif (GrowlerDB.options.auto_wakeup or GrowlerDB.options.auto_call or GrowlerDB.options.auto_growl or GrowlerDB.options.auto_attack or GrowlerDB.options.auto_revive) then
											extramacro = (GrowlerDB.options.auto_call and GrowlerDB.options.activepetslot and GrowlerDB.options.activepetslot ~=0) and extramacro .. "/castsequence [nopet] reset=3 "..GetSpellInfo(l.pets.callpet[GrowlerDB.options.activepetslot])..", "..GetSpellInfo(982).."\n" or extramacro
											extramacro = GrowlerDB.options.auto_revive and extramacro .. "/cast [@pet,dead,exists] "..GetSpellInfo(982).."\n" or extramacro
											extramacro = (GrowlerDB.options.auto_wakeup and spec==1) and extramacro .. "/target Hati\n/cast [@pet,noexists][@pet,dead,exists];[@target,dead,exists] Wake Up\n/targetlasttarget [@target,noharm,exists]\n" or extramacro
											extramacro = GrowlerDB.options.auto_growl and extramacro .. "/click GrowlerBtn_growl\n" or extramacro
											
											if GrowlerDB.options.auto_growl and spec == 3 then
												function string:nl(txt)
													return self .. "\n" .. txt-- .. "\n/g "..txt.."\n/g focus: %f target: %t"
												end
												local function targetmatch(macro,...)
													local targets = {...}
													local t = ""
													t = t:nl("/stopmacro [noexists][dead]")
													t = t:nl("/focus target")
													t = t:nl("/cleartarget")
													for i=1,#targets do
														t = t:nl("/target "..targets[1])
													end
													t = t:nl("/clearfocus [noexists]")
													t = t:nl("/targetlasttarget [noexists]")
													t = t:nl("/stopmacro [@focus,noexists]")
													t = t:nl("/target focus")
													t = t:nl("/targetlasttarget")
													t = t:nl("/target [exists] focus")
													t = t:nl("/clearfocus [exists]")
													t = t:nl("/stopmacro [exists]")
													t = t:nl("/target focus")
													t = t:nl("/clearfocus")
													t = t:nl(macro)
													
													return t
												end
												local growlbtn = CreateFrame("Button","Growler_OverrideBtn_Growl",Growler,"SecureActionButtonTemplate")
												growlbtn:SetAttribute("type","macro")
												growlbtn:SetAttribute("macrotext",targetmatch("/cast [spec:3] Growl","boss1","boss2","boss3","boss4","boss5"))
												extramacro = extramacro .. "/click Growler_OverrideBtn_Growl\n"
											end
											extramacro = GrowlerDB.options.auto_assist and extramacro .. "/petassist\n" or extramacro
											extramacro = GrowlerDB.options.auto_attack and extramacro .. "/petattack [harm,nodead,exists]\n" or extramacro
											extramacro = extramacro .. "/startattack [combat,harm,nodead,exists]\n"
										end
										extramacro = extramacro .. "/console Sound_EnableAllSound 1\n"
										
										local overridebtn = getglobal("Growler_OverrideBtn_"..id) or CreateFrame("Button","Growler_OverrideBtn_"..id,Growler,"SecureActionButtonTemplate")
										overridebtn:SetAttribute("type","macro")
										overridebtn:SetAttribute("macrotext",extramacro..macrobody)
										
										if bind1 then
											SetOverrideBindingClick(GrowlerBar,false,bind1,"Growler_OverrideBtn_"..id,"LeftButton")
										end
										if bind2 then
											SetOverrideBindingClick(GrowlerBar,false,bind2,"Growler_OverrideBtn_"..id,"LeftButton")
										end
									end
								end
							until true
						end
					end
				end
			elseif InCombatLockdown() then
				Growler.msg("You cannot change keybinds in combat.",true)
			end
		end
		Growler.TalentOrButtonChanged = false
		Growler.RefreshKeybinds = false
	end
	
	Growler.DisplayNews = function()
		local news = {
			"v1.52 has introduced autocasting. It makes you do some Growler actions just by using your normal spells. Some are on by default - be sure to check settings to toggle which ones you want.",
			"You can also turn off the alert for some of the automatic spells if you want to save space - they will still fire.",
		}
		for i=1,#news do
			Growler.msg(news[i],false,true)
		end
		Growler.msg(L["You can disable these messages from the options panel."],false,true)
		-- if GrowlerDB.options.sound then PlaySoundKitID(7355) end
	end
end

function Growler_OnLoad(self, event,...)
    self:RegisterEvent("ADDON_LOADED")
end
function Growler_OnEvent(self, event, ...)
	if event == "ADDON_LOADED" then
		if ... == "Growler" then
			self:UnregisterEvent("ADDON_LOADED")
			self:RegisterEvent("PLAYER_ENTERING_WORLD")
			self:RegisterEvent("PLAYER_REGEN_ENABLED")
			self:RegisterEvent("PLAYER_REGEN_DISABLED")
			self:RegisterEvent("CHAT_MSG_ADDON")
			self:RegisterEvent("LOADING_SCREEN_ENABLED")
			self:RegisterEvent("LOADING_SCREEN_DISABLED")
			RegisterAddonMessagePrefix("growler")
			Growler_InitiateDB()
			Growler_Setup()
			
			local locale = GetLocale()
			if locale == "enUS" or locale == "enGB" or locale == "esES" or locale == "esMX" or locale == "deDE" or locale == "frFR" or locale == "itIT" or locale == "ptBR" or locale == "ruRU" then
				if isHunter then
					self:RegisterEvent("PET_STABLE_SHOW")
					self:RegisterEvent("ACTIONBAR_HIDEGRID")
					self:RegisterEvent("ACTIONBAR_PAGE_CHANGED")
					self:RegisterEvent("UPDATE_BONUS_ACTIONBAR")
					self:RegisterEvent("UPDATE_MACROS")
					self:RegisterEvent("PLAYER_TALENT_UPDATE")
					self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
					
					Growler.SetActivePetSlot(true)
					Growler.CreateBar()
					Growler.CreatePetButtons()
					Growler.CreateSpecButtons()
					Growler.CreateStanceButtons()
					Growler.CreateButtons()
					Growler.CreatePortrait()
					S.CreateMainUI()
					if not Growler.HasVisitedStables() then
						Growler.msg(L["Visit the stables to use the stable function. (you only have to do this once)"],true)
					end
				end
			else
				print("|cffff0000{{Growler :: Most functionality, except for the group alert message, is not supported by your language yet. I am working on translating the addon. Thank you for your patience.}}")
			end
			
			SL.Init()
			O.CreateOptionsPanel()
			if GrowlerDB.options.displaynews then Growler.DisplayNews() end
		end
	elseif event == "PLAYER_REGEN_DISABLED" then
			self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	elseif event == "PLAYER_REGEN_ENABLED" then
			self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
		Growler.CheckForGroupGrowl(...)
	elseif event == "CHAT_MSG_ADDON" then
		Growler.AddonMsgRecieved(...)
	elseif event == "PLAYER_ENTERING_WORLD" then
		self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("PLAYER_LEAVING_WORLD")
		Growler.EnteredWorld = true
		
		self:SetScript("OnUpdate", Growler.Update)
	elseif event == "PLAYER_LEAVING_WORLD" then
		self:UnregisterEvent("PLAYER_LEAVING_WORLD")
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		Growler.EnteredWorld = false
		self:SetScript("OnUpdate", nil)
	elseif event == "LOADING_SCREEN_ENABLED" then
		self:SetScript("OnUpdate", nil)
	elseif event == "LOADING_SCREEN_DISABLED" then
		self:SetScript("OnUpdate", Growler.Update)
	end
	
	if isHunter then
		if event == "PLAYER_REGEN_DISABLED" then
			self:UnregisterEvent("PLAYER_TALENT_UPDATE")
			self:UnregisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
		elseif event == "PLAYER_REGEN_ENABLED" then
			Growler.RefreshKeybinds = true
			self:RegisterEvent("PLAYER_TALENT_UPDATE")
			self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
		elseif event == "PLAYER_TALENT_UPDATE" or event == "ACTIONBAR_HIDEGRID" or event == "UPDATE_MACROS" or event == "ACTIVE_TALENT_GROUP_CHANGED" or event == "UPDATE_BONUS_ACTIONBAR" or event == "ACTIONBAR_PAGE_CHANGED" then
			Growler.TalentOrButtonChanged = true
			Growler.TalentChangeMsgShown = false
		elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
			Growler.CheckCombatLogEvent(...)
		elseif event == "PET_STABLE_SHOW" or event == "PET_STABLE_UPDATE" then
			S.UpdateStables()
			if event == "PET_STABLE_SHOW" then
				self:UnregisterEvent("PET_STABLE_SHOW")
				self:RegisterEvent("PET_STABLE_UPDATE")
				self:RegisterEvent("PET_STABLE_CLOSE")
			elseif event == "PET_STABLE_CLOSE" then
				self:RegisterEvent("PET_STABLE_SHOW")
				self:UnregisterEvent("PET_STABLE_UPDATE")
				self:UnregisterEvent("PET_STABLE_CLOSE")
			end
		elseif event == "PLAYER_ENTERING_WORLD" then			
			S.HideStable()
			Growler.bar.ResetPos()
		end
	end
end



































































-------------